"""Unit tests for reviewboard.cmdline.rbsite."""

import os
import shutil
import sys
import tempfile
from io import StringIO

import kgb

from reviewboard.cmdline.rbsite import (Command,
                                        InstallCommand,
                                        ManageCommand,
                                        MissingSiteError,
                                        Site,
                                        UpgradeCommand,
                                        parse_options,
                                        validate_site_paths)
from reviewboard.cmdline.utils.console import init_console, uninit_console
from reviewboard.rb_platform import SITELIST_FILE_UNIX
from reviewboard.testing.testcase import TestCase


class BaseRBSiteTestCase(TestCase):
    """Base class for rb-site unit tests.

    This handles setting up some initial site directories for tests and
    redirecting stderr for capture.
    """

    #: Whether the unit tests need site directories to work with.
    needs_sitedirs = False

    @classmethod
    def setUpClass(cls):
        super(BaseRBSiteTestCase, cls).setUpClass()

        init_console(allow_color=False)

        if cls.needs_sitedirs:
            cls._tempdir = tempfile.mkdtemp(prefix='rb-site-')

            cls.sitelist_filename = os.path.join(cls._tempdir, 'sitelist.txt')
            cls.sitedir1 = os.path.join(cls._tempdir, 'site1')
            cls.sitedir2 = os.path.join(cls._tempdir, 'site2')
            cls.invalid_sitedir = os.path.join(cls._tempdir, 'invalid')

            with open(cls.sitelist_filename, 'w') as fp:
                fp.write('%s\n' % cls.sitedir1)
                fp.write('%s\n' % cls.sitedir2)

            os.mkdir(cls.sitedir1, 0o755)
            os.mkdir(os.path.join(cls.sitedir1, 'conf'), 0o755)

            os.mkdir(cls.sitedir2, 0o755)
            os.mkdir(os.path.join(cls.sitedir2, 'conf'), 0o755)
            os.mkdir(os.path.join(cls.sitedir2, 'htdocs'), 0o755)

    @classmethod
    def tearDownClass(cls):
        super(BaseRBSiteTestCase, cls).tearDownClass()

        if cls.needs_sitedirs and cls._tempdir:
            shutil.rmtree(cls._tempdir)
            cls._tempdir = None

        uninit_console()

    def setUp(self):
        super(BaseRBSiteTestCase, self).setUp()

        self._old_stderr = sys.stderr
        self._old_stdout = sys.stdout
        sys.stderr = StringIO()
        sys.stdout = StringIO()

    def tearDown(self):
        sys.stderr = self._old_stderr
        sys.stdout = self._old_stdout
        self._old_stderr = None
        self._old_stdout = None

        super(BaseRBSiteTestCase, self).tearDown()


class CommandTests(BaseRBSiteTestCase):
    """Unit tests for reviewboard.cmdline.rbsite.Command."""

    def setUp(self):
        super(CommandTests, self).setUp()

        self.command = Command()

    def test_get_site_paths_with_string(self):
        """Testing Command.get_site_paths with site_path as string"""
        class Options(object):
            site_path = '/var/www/reviewboard'

        self.assertEqual(self.command.get_site_paths(Options()),
                         ['/var/www/reviewboard'])

    def test_get_site_paths_with_list(self):
        """Testing Command.get_site_paths with site_path as ststring"""
        class Options(object):
            site_path = [
                '/var/www/reviewboard1',
                '/var/www/reviewboard2',
            ]

        self.assertEqual(
            self.command.get_site_paths(Options()),
            [
                '/var/www/reviewboard1',
                '/var/www/reviewboard2',
            ])

    def test_get_site_paths_without_site_path(self):
        """Testing Command.get_site_paths without site_path"""
        class Options(object):
            site_path = None

        self.assertEqual(self.command.get_site_paths(Options()), [])


class UpgradeCommandTests(BaseRBSiteTestCase):
    """Unit tests for reviewboard.cmdline.rbsite.UpgradeCommand."""

    def setUp(self):
        super(UpgradeCommandTests, self).setUp()

        self.command = UpgradeCommand()

    def test_get_site_paths_with_all_sites(self):
        """Testing UpgradeCommand.get_site_paths with all_sites=True"""
        tmpdir = tempfile.mkdtemp(prefix='rbsite-')

        dir1 = os.path.join(tmpdir, 'site1')
        dir2 = os.path.join(tmpdir, 'site2')
        dir3 = os.path.join(tmpdir, 'site3')

        # Create 2 of the 3 site directories. The third will be excluded, since
        # it doesn't exist.
        os.mkdir(dir1, 0o755)
        os.mkdir(dir2, 0o755)

        site_filename = os.path.join(tmpdir, 'sites')

        with open(site_filename, 'w') as fp:
            fp.write('%s\n' % dir1)
            fp.write('%s\n' % dir2)
            fp.write('%s\n' % dir3)

        class Options(object):
            all_sites = True
            sitelist = site_filename

        try:
            self.assertEqual(self.command.get_site_paths(Options()),
                             [dir1, dir2])
        finally:
            shutil.rmtree(tmpdir)

    def test_get_site_paths_with_all_sites_and_empty(self):
        """Testing UpgradeCommand.get_site_paths with all_sites=True and no
        existing sites in sites file
        """
        tmpdir = tempfile.mkdtemp(prefix='rbsite-')

        # Note that we won't be creating these directories.
        dir1 = os.path.join(tmpdir, 'site1')
        dir2 = os.path.join(tmpdir, 'site2')
        dir3 = os.path.join(tmpdir, 'site3')

        site_filename = os.path.join(tmpdir, 'sites')

        with open(site_filename, 'w') as fp:
            fp.write('%s\n' % dir1)
            fp.write('%s\n' % dir2)
            fp.write('%s\n' % dir3)

        class Options(object):
            all_sites = True
            sitelist = site_filename

        expected_message = \
            'No Review Board sites were listed in %s' % site_filename

        try:
            with self.assertRaisesMessage(MissingSiteError, expected_message):
                self.command.get_site_paths(Options())
        finally:
            shutil.rmtree(tmpdir)

    def test_get_site_paths_with_string(self):
        """Testing UpgradeCommand.get_site_paths with site_path as string"""
        class Options(object):
            all_sites = False
            site_path = '/var/www/reviewboard'

        self.assertEqual(self.command.get_site_paths(Options()),
                         ['/var/www/reviewboard'])

    def test_get_site_paths_with_list(self):
        """Testing UpgradeCommand.get_site_paths with site_path as ststring"""
        class Options(object):
            all_sites = False
            site_path = [
                '/var/www/reviewboard1',
                '/var/www/reviewboard2',
            ]

        self.assertEqual(
            self.command.get_site_paths(Options()),
            [
                '/var/www/reviewboard1',
                '/var/www/reviewboard2',
            ])

    def test_get_site_paths_without_site_path(self):
        """Testing UpgradeCommand.get_site_paths without site_path"""
        class Options(object):
            all_sites = False
            site_path = None

        self.assertEqual(self.command.get_site_paths(Options()), [])


class ParseOptionsTests(BaseRBSiteTestCase):
    """Unit tests for reviewboard.cmdline.rbsite.parse_options."""

    needs_sitedirs = True

    def test_with_no_options(self):
        """Testing rb-site parse_options with no options"""
        self.assertIsNone(parse_options([]))

        help_text = sys.stdout.getvalue()
        self.assertTrue(help_text.startswith('usage: rb-site [-h]'))
        self.assertIn('rb-site helps create, upgrade', help_text)

    def test_with_help_only(self):
        """Testing rb-site parse_options with --help only"""
        with self.assertRaises(SystemExit):
            parse_options(['--help'])

        help_text = sys.stdout.getvalue()
        self.assertTrue(help_text.startswith('usage: rb-site [-h]'))
        self.assertIn('rb-site helps create, upgrade', help_text)

    def test_with_help_first(self):
        """Testing rb-site parse_options with --help as first option"""
        with self.assertRaises(SystemExit):
            parse_options(['--help', 'install', self.sitedir1])

        help_text = sys.stdout.getvalue()
        self.assertTrue(help_text.startswith('usage: rb-site [-h]'))
        self.assertIn('rb-site helps create, upgrade', help_text)

    def test_with_help_as_command_option(self):
        """Testing rb-site parse_options with --help as command option"""
        with self.assertRaises(SystemExit):
            parse_options(['install', self.sitedir1, '--help'])

        help_text = sys.stdout.getvalue()
        self.assertTrue(help_text.startswith('usage: rb-site install'))
        self.assertIn('This will guide you through installing', help_text)

    def test_with_install(self):
        """Testing rb-site parse_options with install"""
        sitedir = os.path.join(self.sitedir1, 'newdir')

        result = parse_options(['install', sitedir])

        self.assertIsNotNone(result)
        self.assertIsInstance(result['command'], InstallCommand)
        self.assertEqual(result['site_paths'], [sitedir])

    def test_with_upgrade(self):
        """Testing rb-site parse_options with upgrade"""
        result = parse_options(['upgrade', self.sitedir1])

        self.assertIsNotNone(result)
        self.assertIsInstance(result['command'], UpgradeCommand)
        self.assertEqual(result['site_paths'], [self.sitedir1])

    def test_with_upgrade_with_invalid_site_dir(self):
        """Testing rb-site parse_options with upgrade with invalid site
        directory
        """
        expected_error = (
            'The site directory "%s" does not exist.'
            % self.invalid_sitedir
        )

        with self.assertRaisesMessage(MissingSiteError, expected_error):
            parse_options(['upgrade', self.invalid_sitedir])

    def test_with_upgrade_with_no_site_dir(self):
        """Testing rb-site parse_options with upgrade and no site directory"""
        expected_message = \
            "You'll need to provide a site directory to run this command."

        with self.assertRaisesMessage(MissingSiteError, expected_message):
            parse_options(['upgrade'])

    def test_with_upgrade_with_all_sites_not_stored(self):
        """Testing rb-site parse_options with upgrade and --all-sites but no
        stored site
        """
        expected_message = \
            "No Review Board sites were listed in %s" % SITELIST_FILE_UNIX

        with self.assertRaisesMessage(MissingSiteError, expected_message):
            parse_options(['upgrade', '--all-sites'])

    def test_with_upgrade_with_all_sites_and_stored(self):
        """Testing rb-site parse_options with upgrade and --all-sites and
        sites stored
        """
        result = parse_options(['upgrade', '--all-sites', '--sitelist',
                                self.sitelist_filename])

        self.assertIsNotNone(result)
        self.assertIsInstance(result['command'], UpgradeCommand)
        self.assertEqual(result['site_paths'], [self.sitedir1, self.sitedir2])

    def test_with_manage_with_invalid_site_dir(self):
        """Testing rb-site parse_options with manage with invalid site
        directory
        """
        expected_error = (
            'The site directory "%s" does not exist.'
            % self.invalid_sitedir
        )

        with self.assertRaisesMessage(MissingSiteError, expected_error):
            parse_options(['manage', self.invalid_sitedir, 'my-command'])

    def test_with_manage_with_command(self):
        """Testing rb-site parse_options with manage and command"""
        result = parse_options(['manage', self.sitedir1, 'my-command'])

        self.assertIsNotNone(result)
        self.assertIsInstance(result['command'], ManageCommand)
        self.assertEqual(result['site_paths'], [self.sitedir1])
        self.assertEqual(result['options'].manage_command, ['my-command'])

    def test_with_manage_with_command_and_args(self):
        """Testing rb-site parse_options with manage and command and args"""
        result = parse_options([
            'manage', self.sitedir1, 'my-command', '--arg1', '--arg2',
            'arg3',
        ])

        self.assertIsNotNone(result)
        self.assertIsInstance(result['command'], ManageCommand)
        self.assertEqual(result['site_paths'], [self.sitedir1])
        self.assertEqual(result['options'].manage_command,
                         ['my-command', '--arg1', '--arg2', 'arg3'])

    def test_with_manage_with_command_and_help_arg(self):
        """Testing rb-site parse_options with manage and command and --help
        argument
        """
        result = parse_options([
            'manage', self.sitedir1, 'my-command', '--help'
        ])

        self.assertIsNotNone(result)
        self.assertIsInstance(result['command'], ManageCommand)
        self.assertEqual(result['site_paths'], [self.sitedir1])
        self.assertEqual(result['options'].manage_command,
                         ['my-command', '--help'])

    def test_with_manage_no_command(self):
        """Testing rb-site parse_options with manage and no command"""
        with self.assertRaises(SystemExit):
            parse_options(['manage', self.sitedir1])

        output = sys.stderr.getvalue()
        self.assertTrue(output.startswith('usage: rb-site manage [-h]'))

        self.assertIn(
            'rb-site manage: error: the following arguments are '
            'required: <command> <args>',
            output)

    def test_with_invalid_command(self):
        """Testing rb-site parse_options with invalid command"""
        with self.assertRaises(SystemExit):
            parse_options(['frob', self.sitedir1])

        output = sys.stderr.getvalue()
        self.assertTrue(output.startswith('usage: rb-site [-h]'))
        self.assertIn("invalid choice: 'frob'", output)


class SiteTests(kgb.SpyAgency, BaseRBSiteTestCase):
    """Unit tests for reviewboard.cmdline.rbsite.Site."""

    needs_sitedirs = True

    def test_generate_settings_local_with_mysql(self):
        """Testing Site.generate_settings_local with MySQL"""
        site = Site(install_dir=self.sitedir1,
                    options={})
        site.domain_name = 'reviews.example.com'
        site.db_type = 'mysql'
        site.db_name = 'test-database'
        site.db_user = 'test-user'
        site.db_pass = 'test-pass'
        site.db_host = 'db.example.com'
        site.db_port = 12345
        site.cache_info = 'localhost:1666'
        site.cache_type = 'memcached'
        site.secret_key = \
            'acdef12345acdef123456abcdef123456abcdef12345abcdef12345'

        site.generate_settings_local()

        self._check_settings_local(
            self.sitedir1,
            'DATABASES = {\n'
            '    "default": {\n'
            '        "ENGINE": "django.db.backends.mysql",\n'
            '        "NAME": "test-database",\n'
            '        "USER": "test-user",\n'
            '        "PASSWORD": "test-pass",\n'
            '        "HOST": "db.example.com",\n'
            '        "PORT": 12345\n'
            '    }\n'
            '}\n'
            'CACHES = {\n'
            '    "default": {\n'
            '        "BACKEND": "django.core.cache.backends.memcached'
            '.PyMemcacheCache",\n'
            '        "LOCATION": "localhost:1666"\n'
            '    }\n'
            '}\n'
            'SECRET_KEY = "acdef12345acdef123456abcdef123456abcdef12345'
            'abcdef12345"\n'
            'SITE_ROOT = ""\n'
            'DEBUG = False\n'
            'ALLOWED_HOSTS = [\n'
            '    "reviews.example.com"\n'
            ']\n'
            'HEALTHCHECK_IPS = []\n')

    def test_generate_settings_local_with_postgres(self):
        """Testing Site.generate_settings_local with Postgres"""
        site = Site(install_dir=self.sitedir1,
                    options={})
        site.domain_name = 'reviews.example.com'
        site.db_type = 'postgresql'
        site.db_name = 'test-database'
        site.db_user = 'test-user'
        site.db_pass = 'test-pass'
        site.db_host = 'db.example.com'
        site.db_port = 12345
        site.cache_info = 'localhost:1666'
        site.cache_type = 'memcached'
        site.secret_key = \
            'acdef12345acdef123456abcdef123456abcdef12345abcdef12345'

        site.generate_settings_local()

        self._check_settings_local(
            self.sitedir1,
            'DATABASES = {\n'
            '    "default": {\n'
            '        "ENGINE": "django.db.backends.postgresql",\n'
            '        "NAME": "test-database",\n'
            '        "USER": "test-user",\n'
            '        "PASSWORD": "test-pass",\n'
            '        "HOST": "db.example.com",\n'
            '        "PORT": 12345\n'
            '    }\n'
            '}\n'
            'CACHES = {\n'
            '    "default": {\n'
            '        "BACKEND": "django.core.cache.backends.memcached'
            '.PyMemcacheCache",\n'
            '        "LOCATION": "localhost:1666"\n'
            '    }\n'
            '}\n'
            'SECRET_KEY = "acdef12345acdef123456abcdef123456abcdef12345'
            'abcdef12345"\n'
            'SITE_ROOT = ""\n'
            'DEBUG = False\n'
            'ALLOWED_HOSTS = [\n'
            '    "reviews.example.com"\n'
            ']\n'
            'HEALTHCHECK_IPS = []\n')

    def test_generate_settings_local_with_sqlite(self):
        """Testing Site.generate_settings_local with SQLite"""
        site = Site(install_dir=self.sitedir1,
                    options={})
        site.domain_name = 'reviews.example.com'
        site.db_type = 'sqlite3'
        site.db_name = '/path/to/reviewboard.db'
        site.cache_info = 'localhost:1666'
        site.cache_type = 'memcached'
        site.secret_key = \
            'acdef12345acdef123456abcdef123456abcdef12345abcdef12345'

        site.generate_settings_local()

        self._check_settings_local(
            self.sitedir1,
            'DATABASES = {\n'
            '    "default": {\n'
            '        "ENGINE": "django.db.backends.sqlite3",\n'
            '        "NAME": "/path/to/reviewboard.db"\n'
            '    }\n'
            '}\n'
            'CACHES = {\n'
            '    "default": {\n'
            '        "BACKEND": "django.core.cache.backends.memcached'
            '.PyMemcacheCache",\n'
            '        "LOCATION": "localhost:1666"\n'
            '    }\n'
            '}\n'
            'SECRET_KEY = "acdef12345acdef123456abcdef123456abcdef12345'
            'abcdef12345"\n'
            'SITE_ROOT = ""\n'
            'DEBUG = False\n'
            'ALLOWED_HOSTS = [\n'
            '    "reviews.example.com"\n'
            ']\n'
            'HEALTHCHECK_IPS = []\n')

    def test_get_settings_upgrade_needed_with_legacy_database(self):
        """Testing Site.get_settings_upgrade_needed with legacy
        DATABASE_* settings
        """
        self.assertTrue(self._get_settings_upgrade_needed({
            'CACHES': {
                'forwarded_backend': {
                    'BACKEND': ('django.core.cache.backends.memcached.'
                                'PyMemcacheCache'),
                    'LOCATION': 'localhost:1666',
                },
            },
            'DATABASE_ENGINE': 'mysql',
        }))

    def test_get_settings_upgrade_needed_with_legacy_cache_backend(self):
        """Testing Site.get_settings_upgrade_needed with legacy
        CACHE_BACKEND setting
        """
        self.assertTrue(self._get_settings_upgrade_needed({
            'CACHE_BACKEND': 'memcached://localhost:1666',
            'DATABASES': {
                'forwarded_backend': {
                    'ENGINE': 'django.db.backends.mysql',
                },
            },
        }))

    def test_get_settings_upgrade_needed_with_legacy_database_engine(self):
        """Testing Site.get_settings_upgrade_needed with legacy database
        engine name
        """
        self.assertTrue(self._get_settings_upgrade_needed({
            'CACHES': {
                'forwarded_backend': {
                    'BACKEND': ('django.core.cache.backends.memcached.'
                                'PyMemcacheCache'),
                    'LOCATION': 'localhost:1666',
                },
            },
            'DATABASES': {
                'default': {
                    'ENGINE': 'mysql',
                },
            },
        }))

    def test_get_settings_upgrade_needed_with_legacy_postgresql_psycopg2(self):
        """Testing Site.get_settings_upgrade_needed with legacy
        postgresql_psycopg2 database engine
        """
        self.assertTrue(self._get_settings_upgrade_needed({
            'CACHES': {
                'forwarded_backend': {
                    'BACKEND': ('django.core.cache.backends.memcached.'
                                'PyMemcacheCache'),
                    'LOCATION': 'localhost:1666',
                },
            },
            'DATABASES': {
                'default': {
                    'ENGINE': 'django.db.backends.postgresql_psycopg2',
                },
            },
        }))

    def test_get_settings_upgrade_needed_with_memcached(self):
        """Testing Site.get_settings_upgrade_needed with modern settings"""
        self.assertTrue(self._get_settings_upgrade_needed({
            'CACHES': {
                'forwarded_backend': {
                    'BACKEND': ('django.core.cache.backends.memcached.'
                                'MemcachedCache'),
                    'LOCATION': 'localhost:1666',
                },
            },
            'DATABASES': {
                'default': {
                    'ENGINE': 'django.db.backends.postgresql',
                },
            },
        }))

    def test_get_settings_upgrade_needed_with_modern_settings(self):
        """Testing Site.get_settings_upgrade_needed with modern settings"""
        self.assertFalse(self._get_settings_upgrade_needed({
            'CACHES': {
                'forwarded_backend': {
                    'BACKEND': ('django.core.cache.backends.memcached.'
                                'PyMemcacheCache'),
                    'LOCATION': 'localhost:1666',
                },
            },
            'DATABASES': {
                'default': {
                    'ENGINE': 'django.db.backends.postgresql',
                },
            },
        }))

    def test_get_wsgi_upgrade_needed_with_rb_pre_4(self):
        """Testing Site.get_wsgi_upgrade_needed with pre-RB4 configuration"""
        self.assertTrue(self._get_wsgi_upgrade_needed(
            "import os\n"
            "import sys\n"
            "\n"
            "os.environ['DJANGO_SETTINGS_MODULE'] = 'reviewboard.settings'\n"
            "os.environ['PYTHON_EGG_CACHE'] = '%(sitedir)s/tmp/egg_cache'\n"
            "os.environ['HOME'] = '%(sitedir)s/data'\n"
            "os.environ['CUSTOM'] = 'abc123'\n"
            "os.environ['PATH'] = '/usr/local/bin:%%s'"
            " %% os.environ['PATH']\n"
            "os.environ['PYTHONPATH'] = '%(sitedir)s/conf:%%s'"
            " %% os.environ['PYTHONPATH']\n"
            "\n"
            "sys.path = ['%(sitedir)s/conf'] + sys.path\n"
            "\n"
            "import django.core.handlers.wsgi\n"
            "application = django.core.handlers.wsgi.WSGIHandler()\n"))

    def test_get_wsgi_upgrade_needed_with_rb4_beta(self):
        """Testing Site.get_wsgi_upgrade_needed with RB4 beta configuration"""
        self.assertTrue(self._get_wsgi_upgrade_needed(
            "import os\n"
            "import sys\n"
            "\n"
            "os.environ['DJANGO_SETTINGS_MODULE'] = 'reviewboard.settings'\n"
            "os.environ['PYTHON_EGG_CACHE'] = '%(sitedir)s/tmp/egg_cache'\n"
            "os.environ['HOME'] = '%(sitedir)s/data'\n"
            "os.environ['CUSTOM'] = 'abc123'\n"
            "os.environ['PATH'] = '/usr/local/bin:%%s'"
            " %% os.environ['PATH']\n"
            "os.environ['PYTHONPATH'] = '%(sitedir)s/conf:%%s'"
            " %% os.environ['PYTHONPATH']\n"
            "\n"
            "sys.path = ['%(sitedir)s/conf'] + sys.path\n"
            "\n"
            "from django.core.wsgi import get_wsgi_application\n"
            "application = get_wsgi_application()\n"))

    def test_get_wsgi_upgrade_needed_with_pkg_resources(self):
        """Testing Site.get_wsgi_upgrade_needed with pkg_resources imports"""
        self.assertTrue(self._get_wsgi_upgrade_needed(
            'import __main__\n'
            '__main__.__requires__ = [\'ReviewBoard\']\n'
            'import pkg_resources\n'
            '\n'
            'from reviewboard.wsgi import application\n'))

    def test_get_wsgi_upgrade_needed_with_rb4(self):
        """Testing Site.get_wsgi_upgrade_needed with RB4+ configuration"""
        self.assertTrue(self._get_wsgi_upgrade_needed(
            "import os\n"
            "import sys\n"
            "\n"
            "os.environ['DJANGO_SETTINGS_MODULE'] = 'reviewboard.settings'\n"
            "os.environ['PYTHON_EGG_CACHE'] = '%(sitedir)s/tmp/egg_cache'\n"
            "os.environ['HOME'] = '%(sitedir)s/data'\n"
            "os.environ['CUSTOM'] = 'abc123'\n"
            "os.environ['PATH'] = '/usr/local/bin:%%s'"
            " %% os.environ['PATH']\n"
            "os.environ['PYTHONPATH'] = '%(sitedir)s/conf:%%s'"
            " %% os.environ['PYTHONPATH']\n"
            "\n"
            "sys.path = ['%(sitedir)s/conf'] + sys.path\n"
            "\n"
            "from reviewboard.wsgi import application\n"))

    def test_get_wsgi_upgrade_needed_with_rb505(self):
        """Testing Site.get_wsgi_upgrade_needed with RB5 through 5.0.5
        configuration
        """
        self.assertTrue(self._get_wsgi_upgrade_needed(
            "import os\n"
            "\n"
            "os.environ['REVIEWBOARD_SITEDIR'] = '%(sitedir)s'\n"
            "\n"
            "from reviewboard.wsgi import application\n"))

    def test_get_wsgi_upgrade_needed_with_rb506(self):
        """Testing Site.get_wsgi_upgrade_needed with RB 5.0.6+ configuration"""
        self.assertFalse(self._get_wsgi_upgrade_needed(
            "import os\n"
            "import sys\n"
            "\n"
            "\n"
            "os.environ['REVIEWBOARD_SITEDIR'] = '%(sitedir)s'\n"
            "\n"
            "sys.path.insert(0, os.path.join(\n"
            "    os.environ['REVIEWBOARD_SITEDIR'], 'venv', 'lib',\n"
            "    'python%%s.%%s' %% sys.version_info[:2], 'site-packages'))\n"
            "\n"
            "\n"
            "# BEGIN CUSTOM SETTINGS\n"
            "\n"
            "# END CUSTOM SETTINGS\n"
            "\n"
            "\n"
            "from reviewboard.wsgi import application\n"))

    def test_upgrade_settings_with_legacy_database(self):
        """Testing Site.upgrade_settings with legacy DATABASE_* settings"""
        self._check_upgrade_settings(
            stored_settings={
                'CACHES': {
                    'forwarded_backend': {
                        'BACKEND': ('django.core.cache.backends.memcached.'
                                    'PyMemcacheCache'),
                        'LOCATION': 'localhost:1666',
                    },
                },
                'DATABASE_ENGINE': 'mysql',
                'DATABASE_NAME': 'test-database',
                'DATABASE_HOST': 'db.example.com',
                'DATABASE_PORT': 12345,
                'DATABASE_USER': 'test-user',
                'DATABASE_PASSWORD': 'test-pass',
            },
            stored_settings_text=(
                'DATABASE_ENGINE = "mysql"\n'
                'DATABASE_NAME = "test-database"\n'
                'DATABASE_HOST = "db.example.com"\n'
                'DATABASE_PORT = 12345\n'
                'DATABASE_USER = "test-user"\n'
                'DATABASE_PASSWORD = "test-pass"\n'
                'CACHES = {\n'
                '    "default": {\n'
                '        "BACKEND": "django.core.cache.backends.memcached.'
                'PyMemcacheCache",\n'
                '        "LOCATION": "localhost:1666",\n'
                '    },\n'
                '}\n'
            ),
            expected_settings_text=(
                'DATABASES = {\n'
                '    "default": {\n'
                '        "ENGINE": "django.db.backends.mysql",\n'
                '        "NAME": "test-database",\n'
                '        "USER": "test-user",\n'
                '        "PASSWORD": "test-pass",\n'
                '        "HOST": "db.example.com",\n'
                '        "PORT": 12345\n'
                '    }\n'
                '}\n'
                'CACHES = {\n'
                '    "default": {\n'
                '        "BACKEND": "django.core.cache.backends.memcached.'
                'PyMemcacheCache",\n'
                '        "LOCATION": "localhost:1666",\n'
                '    },\n'
                '}\n'
            ))

    def test_upgrade_settings_with_legacy_cache_backend(self):
        """Testing Site.upgrade_settings with legacy CACHE_BACKEND setting"""
        self._check_upgrade_settings(
            stored_settings={
                'CACHE_BACKEND': 'memcached://localhost:1666',
                'DATABASES': {
                    'default': {
                        'ENGINE': 'mysql',
                    },
                },
            },
            stored_settings_text=(
                'CACHE_BACKEND = "memcached://localhost:1666"\n'
                'DATABASES = {\n'
                '    "default": {\n'
                '        "ENGINE": "django.db.backends.mysql",\n'
                '    },\n'
                '}\n'
            ),
            expected_settings_text=(
                'CACHES = {\n'
                '    "default": {\n'
                '        "BACKEND": "django.core.cache.backends.memcached.'
                'PyMemcacheCache",\n'
                '        "LOCATION": "localhost:1666"\n'
                '    }\n'
                '}\n'
                'DATABASES = {\n'
                '    "default": {\n'
                '        "ENGINE": "django.db.backends.mysql",\n'
                '    },\n'
                '}\n'
            ))

    def test_upgrade_settings_with_legacy_database_engine(self):
        """Testing Site.upgrade_settings with legacy database engine name"""
        self._check_upgrade_settings(
            stored_settings={
                'CACHES': {
                    'forwarded_backend': {
                        'BACKEND': ('django.core.cache.backends.memcached.'
                                    'PyMemcacheCache'),
                        'LOCATION': 'localhost:1666',
                    },
                },
                'DATABASES': {
                    'default': {
                        'ENGINE': 'mysql',
                    },
                },
            },
            stored_settings_text=(
                'CACHES = {\n'
                '    "default": {\n'
                '        "BACKEND": "django.core.cache.backends.memcached.'
                'PyMemcacheCache",\n'
                '        "LOCATION": "localhost:1666",\n'
                '    },\n'
                '}\n'
                'DATABASES = {\n'
                '    "default": {\n'
                '        "ENGINE": "mysql",\n'
                '    },\n'
                '}\n'
            ),
            expected_settings_text=(
                'CACHES = {\n'
                '    "default": {\n'
                '        "BACKEND": "django.core.cache.backends.memcached.'
                'PyMemcacheCache",\n'
                '        "LOCATION": "localhost:1666",\n'
                '    },\n'
                '}\n'
                'DATABASES = {\n'
                '    "default": {\n'
                '        "ENGINE": "django.db.backends.mysql",\n'
                '    },\n'
                '}\n'
            ))

    def test_upgrade_settings_with_legacy_postgresql_psycopg2(self):
        """Testing Site.upgrade_settings with legacy postgresql_psycopg2
        database engine
        """
        self._check_upgrade_settings(
            stored_settings={
                'CACHES': {
                    'forwarded_backend': {
                        'BACKEND': ('django.core.cache.backends.memcached.'
                                    'PyMemcacheCache'),
                        'LOCATION': 'localhost:1666',
                    },
                },
                'DATABASES': {
                    'default': {
                        'ENGINE': 'django.db.backends.postgresql_psycopg2',
                    },
                },
            },
            stored_settings_text=(
                'CACHES = {\n'
                '    "default": {\n'
                '        "BACKEND": "django.core.cache.backends.memcached.'
                'PyMemcacheCache",\n'
                '        "LOCATION": "localhost:1666",\n'
                '    },\n'
                '}\n'
                'DATABASES = {\n'
                '    "default": {\n'
                '        "ENGINE": "django.db.backends.postgresql_psycopg2",\n'
                '    },\n'
                '}\n'
            ),
            expected_settings_text=(
                'CACHES = {\n'
                '    "default": {\n'
                '        "BACKEND": "django.core.cache.backends.memcached.'
                'PyMemcacheCache",\n'
                '        "LOCATION": "localhost:1666",\n'
                '    },\n'
                '}\n'
                'DATABASES = {\n'
                '    "default": {\n'
                '        "ENGINE": "django.db.backends.postgresql",\n'
                '    },\n'
                '}\n'
            ))

    def test_upgrade_settings_with_modern_settings(self):
        """Testing Site.upgrade_settings with modern settings"""
        self._check_upgrade_settings(
            stored_settings={
                'CACHES': {
                    'forwarded_backend': {
                        'BACKEND': ('django.core.cache.backends.memcached.'
                                    'PyMemcacheCache'),
                        'LOCATION': 'localhost:1666',
                    },
                },
                'DATABASES': {
                    'default': {
                        'ENGINE': 'django.db.backends.postgresql',
                    },
                },
            },
            stored_settings_text=(
                'CACHES = {\n'
                '    "default": {\n'
                '        "BACKEND": "django.core.cache.backends.memcached.'
                'PyMemcacheCache",\n'
                '        "LOCATION": "localhost:1666",\n'
                '    },\n'
                '}\n'
                'DATABASES = {\n'
                '    "default": {\n'
                '        "ENGINE": "django.db.backends.postgresql",\n'
                '    },\n'
                '}\n'
            ),
            expected_settings_text=(
                'CACHES = {\n'
                '    "default": {\n'
                '        "BACKEND": "django.core.cache.backends.memcached.'
                'PyMemcacheCache",\n'
                '        "LOCATION": "localhost:1666",\n'
                '    },\n'
                '}\n'
                'DATABASES = {\n'
                '    "default": {\n'
                '        "ENGINE": "django.db.backends.postgresql",\n'
                '    },\n'
                '}\n'
            ))

    def test_upgrade_wsgi_with_rb_pre_4(self):
        """Testing Site.upgrade_wsgi with pre-RB4 configuration"""
        self._check_upgrade_wsgi(
            ("import __main__\n"
             "__main__.__requires__ = ['ReviewBoard']\n"
             "import pkg_resources\n"
             "\n"
             "import os\n"
             "import sys\n"
             "\n"
             "os.environ['DJANGO_SETTINGS_MODULE'] = 'reviewboard.settings'\n"
             "os.environ['PYTHON_EGG_CACHE'] = '%(sitedir)s/tmp/egg_cache'\n"
             "os.environ['HOME'] = '%(sitedir)s/data'\n"
             "os.environ['CUSTOM'] = 'abc123'\n"
             "os.environ['PATH'] = '/usr/local/bin:%%s'"
             " %% os.environ['PATH']\n"
             "os.environ['PYTHONPATH'] = '%(sitedir)s/conf:%%s'"
             " %% os.environ['PYTHONPATH']\n"
             "\n"
             "sys.path = ['%(sitedir)s/conf'] + sys.path\n"
             "\n"
             "import django.core.handlers.wsgi\n"
             "application = django.core.handlers.wsgi.WSGIHandler()\n"),
            ("import os\n"
             "import sys\n"
             "\n"
             "\n"
             "os.environ['REVIEWBOARD_SITEDIR'] = '%(sitedir)s'\n"
             "\n"
             "sys.path.insert(0, os.path.join(\n"
             "    os.environ['REVIEWBOARD_SITEDIR'], 'venv', 'lib',\n"
             "    'python%%s.%%s' %% sys.version_info[:2], 'site-packages'))\n"
             "\n"
             "\n"
             "# BEGIN CUSTOM SETTINGS\n"
             "os.environ['CUSTOM'] = 'abc123'\n"
             "os.environ['PATH'] = '/usr/local/bin:%%s'"
             " %% os.environ['PATH']\n"
             "# END CUSTOM SETTINGS\n"
             "\n"
             "\n"
             "from reviewboard.wsgi import application\n"))

    def test_upgrade_wsgi_with_rb_4_beta(self):
        """Testing Site.upgrade_wsgi with RB4 beta configuration"""
        self._check_upgrade_wsgi(
            ("import __main__\n"
             "__main__.__requires__ = ['ReviewBoard']\n"
             "import pkg_resources\n"
             "\n"
             "import os\n"
             "import sys\n"
             "\n"
             "os.environ['DJANGO_SETTINGS_MODULE'] = 'reviewboard.settings'\n"
             "os.environ['PYTHON_EGG_CACHE'] = '%(sitedir)s/tmp/egg_cache'\n"
             "os.environ['HOME'] = '%(sitedir)s/data'\n"
             "os.environ['CUSTOM'] = 'abc123'\n"
             "os.environ['PATH'] = '/usr/local/bin:%%s'"
             " %% os.environ['PATH']\n"
             "os.environ['PYTHONPATH'] = '%(sitedir)s/conf:%%s'"
             " %% os.environ['PYTHONPATH']\n"
             "\n"
             "sys.path = ['%(sitedir)s/conf'] + sys.path\n"
             "\n"
             "from django.core.wsgi import get_wsgi_application\n"
             "application = get_wsgi_application()\n"),
            ("import os\n"
             "import sys\n"
             "\n"
             "\n"
             "os.environ['REVIEWBOARD_SITEDIR'] = '%(sitedir)s'\n"
             "\n"
             "sys.path.insert(0, os.path.join(\n"
             "    os.environ['REVIEWBOARD_SITEDIR'], 'venv', 'lib',\n"
             "    'python%%s.%%s' %% sys.version_info[:2], 'site-packages'))\n"
             "\n"
             "\n"
             "# BEGIN CUSTOM SETTINGS\n"
             "os.environ['CUSTOM'] = 'abc123'\n"
             "os.environ['PATH'] = '/usr/local/bin:%%s'"
             " %% os.environ['PATH']\n"
             "# END CUSTOM SETTINGS\n"
             "\n"
             "\n"
             "from reviewboard.wsgi import application\n"))

    def test_upgrade_wsgi_with_rb_5_0_5(self):
        """Testing Site.upgrade_wsgi with RB 5.0-5.0.5 configuration"""
        self._check_upgrade_wsgi(
            ("import os\n"
             "\n"
             "os.environ['REVIEWBOARD_SITEDIR'] = '%(sitedir)s'\n"
             "\n"
             "from reviewboard.wsgi import application\n"),
            ("import os\n"
             "import sys\n"
             "\n"
             "\n"
             "os.environ['REVIEWBOARD_SITEDIR'] = '%(sitedir)s'\n"
             "\n"
             "sys.path.insert(0, os.path.join(\n"
             "    os.environ['REVIEWBOARD_SITEDIR'], 'venv', 'lib',\n"
             "    'python%%s.%%s' %% sys.version_info[:2], 'site-packages'))\n"
             "\n"
             "\n"
             "# BEGIN CUSTOM SETTINGS\n"
             "\n"
             "# END CUSTOM SETTINGS\n"
             "\n"
             "\n"
             "from reviewboard.wsgi import application\n"))

    def test_upgrade_wsgi_with_rb_5_0_6(self):
        """Testing Site.upgrade_wsgi with RB 5.0.6+ configuration"""
        self._check_upgrade_wsgi(
            ("import os\n"
             "\n"
             "os.environ['REVIEWBOARD_SITEDIR'] = '%(sitedir)s'\n"
             "\n"
             "sys.path.insert(0, os.path.join(\n"
             "    os.environ['REVIEWBOARD_SITEDIR'], 'venv', 'lib',\n"
             "    'python%%s.%%s' %% sys.version_info[:2], 'site-packages'))\n"
             "\n"
             "\n"
             "# BEGIN CUSTOM SETTINGS\n"
             "\n"
             "# END CUSTOM SETTINGS\n"
             "\n"
             "\n"
             "from reviewboard.wsgi import application\n"),
            ("import os\n"
             "import sys\n"
             "\n"
             "\n"
             "os.environ['REVIEWBOARD_SITEDIR'] = '%(sitedir)s'\n"
             "\n"
             "sys.path.insert(0, os.path.join(\n"
             "    os.environ['REVIEWBOARD_SITEDIR'], 'venv', 'lib',\n"
             "    'python%%s.%%s' %% sys.version_info[:2], 'site-packages'))\n"
             "\n"
             "\n"
             "# BEGIN CUSTOM SETTINGS\n"
             "\n"
             "# END CUSTOM SETTINGS\n"
             "\n"
             "\n"
             "from reviewboard.wsgi import application\n"))

    def test_upgrade_wsgi_with_rb_5_0_6_custom(self):
        """Testing Site.upgrade_wsgi with RB 5.0.6+ configuration with
        custom settings
        """
        self._check_upgrade_wsgi(
            ("import os\n"
             "\n"
             "os.environ['REVIEWBOARD_SITEDIR'] = '%(sitedir)s'\n"
             "\n"
             "sys.path.insert(0, os.path.join(\n"
             "    os.environ['REVIEWBOARD_SITEDIR'], 'venv', 'lib',\n"
             "    'python%%s.%%s' %% sys.version_info[:2], 'site-packages'))\n"
             "\n"
             "\n"
             "# BEGIN CUSTOM SETTINGS\n"
             "\n"
             "# Some comment.\n"
             "import foo\n"
             "os.environ['FOO'] = foo.get_bar()\n"
             "\n"
             "# END CUSTOM SETTINGS\n"
             "\n"
             "\n"
             "from reviewboard.wsgi import application\n"),
            ("import os\n"
             "import sys\n"
             "\n"
             "\n"
             "os.environ['REVIEWBOARD_SITEDIR'] = '%(sitedir)s'\n"
             "\n"
             "sys.path.insert(0, os.path.join(\n"
             "    os.environ['REVIEWBOARD_SITEDIR'], 'venv', 'lib',\n"
             "    'python%%s.%%s' %% sys.version_info[:2], 'site-packages'))\n"
             "\n"
             "\n"
             "# BEGIN CUSTOM SETTINGS\n"
             "\n"
             "# Some comment.\n"
             "import foo\n"
             "os.environ['FOO'] = foo.get_bar()\n"
             "\n"
             "# END CUSTOM SETTINGS\n"
             "\n"
             "\n"
             "from reviewboard.wsgi import application\n"))

    def test_upgrade_settings_with_legacy_memcached_backend(self):
        """Testing Site.upgrade_settings with legacy MemcachedCache cache
        backend
        """
        self._check_upgrade_settings(
            stored_settings={
                'CACHES': {
                    'forwarded_backend': {
                        'BACKEND': ('django.core.cache.backends.memcached.'
                                    'MemcachedCache'),
                        'LOCATION': 'localhost:1666',
                    },
                },
                'DATABASES': {
                    'default': {
                        'ENGINE': 'django.db.backends.postgresql',
                    },
                },
            },
            stored_settings_text=(
                'CACHES = {\n'
                '    "default": {\n'
                '        "BACKEND": "django.core.cache.backends.memcached.'
                'MemcachedCache",\n'
                '        "LOCATION": "localhost:1666",\n'
                '    },\n'
                '}\n'
                'DATABASES = {\n'
                '    "default": {\n'
                '        "ENGINE": "django.db.backends.postgresql",\n'
                '    },\n'
                '}\n'
            ),
            expected_settings_text=(
                'CACHES = {\n'
                '    "default": {\n'
                '        "BACKEND": "django.core.cache.backends.memcached.'
                'PyMemcacheCache",\n'
                '        "LOCATION": "localhost:1666",\n'
                '    },\n'
                '}\n'
                'DATABASES = {\n'
                '    "default": {\n'
                '        "ENGINE": "django.db.backends.postgresql",\n'
                '    },\n'
                '}\n'
            ))

    def _get_settings_upgrade_needed(self, stored_settings):
        """Return Site.get_settings_upgrade_needed with the provided settings.

        Args:
            stored_settings (dict):
                A dictionary of settings that would be stored in
                :file:`settings_local.py`.

        Returns:
            bool:
            The result of :py:class:`~reviewboard.cmdline.rbsite.Site.
            get_settings_upgrade_needed`.
        """
        class SettingsLocal(object):
            pass

        for key, value in stored_settings.items():
            setattr(SettingsLocal, key, value)

        site = Site(install_dir=self.sitedir2,
                    options={})
        self.spy_on(site.get_settings_local,
                    op=kgb.SpyOpReturn(SettingsLocal))

        return site.get_settings_upgrade_needed()

    def _get_wsgi_upgrade_needed(self, stored_wsgi_text):
        """Return Site.get_wsgi_upgrade_needed with the provided file.

        Args:
            stored_settings (dict):
                A dictionary of settings that would be stored in
                :file:`settings_local.py`.

        Returns:
            bool:
            The result of :py:class:`~reviewboard.cmdline.rbsite.Site.
            get_settings_upgrade_needed`.
        """
        sitedir = self.sitedir2
        site = Site(install_dir=sitedir,
                    options={})

        filename = os.path.join(site.abs_install_dir, 'htdocs',
                                'reviewboard.wsgi')

        with open(filename, 'w') as fp:
            fp.write(stored_wsgi_text % {
                'sitedir': sitedir,
            })

        return site.get_wsgi_upgrade_needed()

    def _check_upgrade_settings(self, stored_settings, stored_settings_text,
                                expected_settings_text):
        """Check that upgrading settings produces the expected results.

        Args:
            stored_settings (dict):
                A dictionary of settings that would be stored in
                :file:`settings_local.py`.

            stored_settings_text (unicode):
                The content of the :file:`settings_local.py` file to write
                and upgrade.

            expected_settings_text (unicode):
                The content of just the settings, without any blank lines
                or comments.

        Raises:
            AssertionError:
                An expectation failed.
        """
        class SettingsLocal(object):
            pass

        for key, value in stored_settings.items():
            setattr(SettingsLocal, key, value)

        site = Site(install_dir=self.sitedir2,
                    options={})
        self.spy_on(site.get_settings_local,
                    op=kgb.SpyOpReturn(SettingsLocal))

        with open(os.path.join(site.abs_install_dir, 'conf',
                               'settings_local.py'),
                  'w') as fp:
            fp.write(stored_settings_text)

        site.upgrade_settings()

        self._check_settings_local(site.abs_install_dir,
                                   expected_settings_text)

    def _check_upgrade_wsgi(self, stored_wsgi_text, expected_wsgi_text):
        """Check that upgrading reviewboard.wsgi produces the expected results.

        Args:
            stored_wsgi_text (unicode):
                The contents of the :file:`htdocs/reviewboard.wsgi` file to
                write and upgrade.

            expected_wsgi_text (unicode):
                The expected content of the file.

        Raises:
            AssertionError:
                An expectation failed.
        """
        sitedir = self.sitedir2
        site = Site(install_dir=sitedir,
                    options={})

        filename = os.path.join(site.abs_install_dir, 'htdocs',
                                'reviewboard.wsgi')

        with open(filename, 'w') as fp:
            fp.write(stored_wsgi_text % {
                'sitedir': sitedir,
            })

        site.upgrade_wsgi()

        with open(filename, 'r') as fp:
            self.assertMultiLineEqual(
                fp.read(),
                expected_wsgi_text % {
                    'sitedir': sitedir,
                })

    def _check_settings_local(self, sitedir, expected_settings_text):
        """Check that a generated settings_local.py has the expected settings.

        This will verify the existence of the file in the site directory,
        strip away any comments or blank lines, and compare against the
        expected settings text.

        Args:
            sitedir (unicode):
                The path to the site directory.

            expected_settings_text (unicode):
                The content of just the settings, without any blank lines
                or comments.

        Raises:
            AssertionError:
                An expectation failed.
        """
        filename = os.path.join(sitedir, 'conf', 'settings_local.py')
        self.assertTrue(os.path.exists(filename))

        with open(filename, 'r') as fp:
            lines = fp.readlines()

        # Strip away all comments and blank lines.
        lines = [
            line
            for line in lines
            if line.strip() and not line.startswith('#')
        ]

        self.assertMultiLineEqual(''.join(lines), expected_settings_text)


class ValidateSitePathsTests(BaseRBSiteTestCase):
    """Unit tests for reviewboard.cmdline.rbsite.validate_site_paths."""

    def test_with_valid_sites(self):
        """Testing validate_site_paths with valid sites"""
        # This should not raise.
        validate_site_paths([os.path.dirname(__file__)])

    def test_with_empty(self):
        """Testing validate_site_paths with empty list"""
        expected_message = \
            "You'll need to provide a site directory to run this command."

        with self.assertRaisesMessage(MissingSiteError, expected_message):
            validate_site_paths([])

        with self.assertRaisesMessage(MissingSiteError, expected_message):
            validate_site_paths(None)

    def test_with_missing_site(self):
        """Testing validate_site_paths with missing site"""
        expected_message = 'The site directory "/test" does not exist.'

        with self.assertRaisesMessage(MissingSiteError, expected_message):
            validate_site_paths(['/test'])

    def test_with_missing_site_and_require_exists_false(self):
        """Testing validate_site_paths with missing site and
        require_exists=False
        """
        # This should not raise.
        validate_site_paths(['/test'], require_exists=False)



"""View mixins provided by the admin app."""

from django.utils.decorators import method_decorator

from reviewboard.admin.decorators import check_read_only


class CheckReadOnlyViewMixin(object):
    """View mixin to check if the site is read-only.

    This is a convenience around using the :py:func:`@check_reead_only
    <reviewboard.admin.decorators.check_read_only>` decorator for class-based
    views.
    """

    @method_decorator(check_read_only)
    def dispatch(self, *args, **kwargs):
        """Dispatch a HTTP request to the right handler.

        Args:
            *args (tuple):
                Positional arguments to pass to the handler.

            **kwargs (tuple):
                Keyword arguments to pass to the handler.

                These will be arguments provided by the URL pattern.

        Returns:
            django.http.HttpResponse:
            The resulting HTTP response to send to the client.
        """
        return super(CheckReadOnlyViewMixin, self).dispatch(*args, **kwargs)



"""Default UI themes for Review Board.

Version Added:
    7.0
"""

from __future__ import annotations

from django.utils.translation import gettext_lazy as _

from reviewboard.themes.ui.base import BaseUITheme, InkColorScheme


class LightUITheme(BaseUITheme):
    """A light mode theme.

    This is the traditional light appearance used for Review Board.

    Version Added:
        7.0
    """

    theme_id = 'light'
    name = _('Light mode')
    ink_color_scheme = InkColorScheme.LIGHT


class DarkUITheme(BaseUITheme):
    """A dark mode theme.

    Version Added:
        7.0
    """

    theme_id = 'dark'
    name = _('Dark mode')
    ink_color_scheme = InkColorScheme.DARK


class SystemUITheme(BaseUITheme):
    """A theme using the current system appearance.

    This will auto-select light mode or dark mode based on the
    browser-specified preference.

    Version Added:
        7.0
    """

    theme_id = 'system'
    name = _('System theme')
    ink_color_scheme = InkColorScheme.SYSTEM



"""Base support for file attachment unit tests.

Version Added:
    7.0.3:
    This was split off from :py:mod:`reviewboard.attachments.tests`.
"""

from __future__ import annotations

import os
from typing import TYPE_CHECKING

from django.conf import settings
from django.contrib.auth.models import User
from django.core.files.uploadedfile import SimpleUploadedFile

from reviewboard.attachments.models import FileAttachment
from reviewboard.diffviewer.models import DiffSet, FileDiff
from reviewboard.scmtools.core import PRE_CREATION
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from reviewboard.diffviewer.models import DiffSetHistory
    from reviewboard.scmtools.core import Revision


class BaseFileAttachmentTestCase(TestCase):
    """Base support for file attachment unit tests.

    Version Added:
        7.0.3:
        This was split off from :py:mod:`reviewboard.attachments.tests`.
    """

    fixtures = ['test_users', 'test_scmtools', 'test_site']

    DEFINED_LOCAL_SITE_KEY = FileAttachment.DEFINED_LOCAL_SITE_KEY

    def make_uploaded_file(self) -> SimpleUploadedFile:
        """Create a return a file to use for mocking in forms."""
        filename = os.path.join(settings.STATIC_ROOT,
                                'rb', 'images', 'logo.png')

        with open(filename, 'rb') as fp:
            uploaded_file = SimpleUploadedFile(fp.name, fp.read(),
                                               content_type='image/png')

        return uploaded_file

    def make_filediff(
        self,
        *,
        is_new: bool = False,
        diffset_history: (DiffSetHistory | None) = None,
        diffset_revision: int = 1,
        source_filename: str = 'file1',
        dest_filename: str = 'file2',
        with_local_site: bool = False,
    ) -> FileDiff:
        """Create and return a FileDiff with the given data.

        Version Changed:
            8.1:
            Made arguments keyword only and added ``with_local_site``.

        Args:
            is_new (bool, optional):
                Whether the FileDiff is new.

            diffset_history (reviewboard.diffviewer.models.DiffSetHistory,
                             optional):
                The optional diffset history to set.

            diffset_revision (int, optional)):
                The revision of the diffset.

            source_filename (str, optional)):
                The name of the source file.

            dest_filename (str, optional)):
                The name of the destination file.

            with_local_site (bool, optional):
                Whether to create the repository using a Local Site. This
                will choose one based on :py:attr:`local_site_name`.

                Version Added:
                    8.1
        """
        source_revision: Revision | str
        dest_revision: Revision | str

        if is_new:
            source_revision = PRE_CREATION
            dest_revision = ''
        else:
            source_revision = '1'
            dest_revision = '2'

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        repository = self.create_repository(local_site=local_site)

        if not diffset_history:
            user = User.objects.get(username='doc')
            review_request = self.create_review_request(repository=repository,
                                                        submitter=user,
                                                        local_site=local_site)
            diffset_history = review_request.diffset_history

        diffset = DiffSet.objects.create(name='test',
                                         revision=diffset_revision,
                                         repository=repository,
                                         history=diffset_history)
        filediff = FileDiff(source_file=source_filename,
                            source_revision=source_revision,
                            dest_file=dest_filename,
                            dest_detail=dest_revision,
                            diffset=diffset,
                            binary=True)
        filediff.save()

        return filediff



"""Added Profile.extra_data.

Version Added:
    1.7.8
"""

from django_evolution.mutations import AddField
from djblets.db.fields import JSONField


MUTATIONS = [
    AddField('Profile', 'extra_data', JSONField, null=True)
]



"""Server information and capability registration for the API."""

import logging
from copy import deepcopy

from django.conf import settings
from djblets.siteconfig.models import SiteConfiguration

from reviewboard import get_version_string, get_package_version, is_release
from reviewboard.admin.server import get_server_url
from reviewboard.diffviewer.features import dvcs_feature
from reviewboard.reviews.ui import review_ui_registry
from reviewboard.scmtools import scmtools_registry


logger = logging.getLogger(__name__)


_registered_capabilities = {}
_capabilities_defaults = {
    'authentication': {
        # Whether to allow clients to authenticate to Review Board
        # via a web browser.
        'client_web_login': True,
    },
    'diffs': {
        'base_commit_ids': True,
        'file_attachments': True,
        'moved_files': True,
        'validation': {
            'base_commit_ids': True,
        }
    },
    'extra_data': {
        'json_patching': True,
    },
    'review_requests': {
        'commit_ids': True,
        'trivial_publish': True,
    },
    'review_uis': {
        'supported_mimetypes': [],
    },
    'scmtools': {
        'git': {
            'empty_files': True,
            'symlinks': True,
        },
        'mercurial': {
            'empty_files': True,
        },
        'perforce': {
            'moved_files': True,
            'empty_files': True,
        },
        'svn': {
            'empty_files': True,
        },
    },
    'text': {
        'markdown': True,
        'per_field_text_types': True,
        'can_include_raw_values': True,
    },
}
_feature_gated_capabilities = {
    'review_requests': {
        'supports_history': dvcs_feature,
    },
}


def get_server_info(request=None):
    """Return server information for use in the API.

    This is used for the root resource and for the deprecated server
    info resource.

    Args:
        request (django.http.HttpRequest, optional):
            The HTTP request from the client.

    Returns:
        dict:
        A dictionary of information about the server and its capabilities.
    """
    return {
        'product': {
            'name': 'Review Board',
            'version': get_version_string(),
            'package_version': get_package_version(),
            'is_release': is_release(),
        },
        'site': {
            'url': get_server_url(request=request),
            'administrators': [
                {
                    'name': name,
                    'email': email,
                }
                for name, email in settings.ADMINS
            ],
            'time_zone': settings.TIME_ZONE,
        },
        'capabilities': get_capabilities(request=request),
    }


def get_capabilities(request=None):
    """Return the capabilities made available in the API.

    Args:
        request (django.http.HttpRequest, optional):
            The http request from the client.

    Returns:
        dict:
        The dictionary of capabilities.
    """
    capabilities = deepcopy(_capabilities_defaults)
    capabilities.update(_registered_capabilities)

    for category, cap, enabled in get_feature_gated_capabilities(request):
        capabilities.setdefault(category, {})[cap] = enabled

    siteconfig = SiteConfiguration.objects.get_current()
    capabilities['authentication']['client_web_login'] = \
        siteconfig.get('client_web_login')

    capabilities['diffs'].update({
        'max_binary_size': siteconfig.get('diffviewer_max_binary_size'),
        'max_diff_size': siteconfig.get('diffviewer_max_diff_size'),
    })

    # We always report support for Power Pack-provided types, so that clients
    # can upload these files, even if Power Pack is not (yet) licensed.
    mimetypes = {
        'application/msword',
        'application/pdf',
        'application/vnd.ms-excel',
        'application/vnd.ms-powerpoint',
        'application/vnd.oasis.opendocument.presentation',
        'application/vnd.oasis.opendocument.spreadsheet',
        'application/vnd.oasis.opendocument.text',
        'application/vnd.openxmlformats-officedocument.presentationml.presentation',
        'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
        'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
        'application/x-pdf',
    }

    for review_ui_class in review_ui_registry:
        mimetypes.update(review_ui_class.supported_mimetypes)

    capabilities['review_uis']['supported_mimetypes'] = sorted(mimetypes)
    capabilities['scmtools']['supported_tools'] = sorted(
        scmtool.scmtool_id
        for scmtool in scmtools_registry
    )

    return capabilities


def get_feature_gated_capabilities(request=None):
    """Return the capabilities gated behind enabled features.

    Args:
        request (django.http.HttpRequest, optional):
            The HTTP request from the client.

    Yields:
        tuple:
        A 3-tuple of the following:

        * The category of the capability (:py:class:`unicode`).
        * The capability name (:py:class:`unicode`).
        * Whether or not the capability is enabled (:py:class:`bool`).
    """
    for category, caps in _feature_gated_capabilities.items():
        for cap, required_feature in caps.items():
            if required_feature.is_enabled(request=request):
                yield category, cap, True


def register_webapi_capabilities(capabilities_id, caps):
    """Register a set of web API capabilities.

    These capabilities will appear in the dictionary of available
    capabilities with the ID as their key.

    A capabilities_id attribute passed in, and can only be registered once.
    A KeyError will be thrown if attempting to register a second time.

    Args:
        capabilities_id (unicode):
            A unique ID representing this collection of capabilities.
            This can only be used once until unregistered.

        caps (dict):
            The dictionary of capabilities to register. Each key msut
            be a string, and each value should be a boolean or a
            dictionary of string keys to booleans.

    Raises:
        KeyError:
            The capabilities ID has already been used.
    """
    if not capabilities_id:
        raise ValueError('The capabilities_id attribute must not be None')

    if capabilities_id in _registered_capabilities:
        raise KeyError('"%s" is already a registered set of capabilities'
                       % capabilities_id)

    if capabilities_id in _capabilities_defaults:
        raise KeyError('"%s" is reserved for the default set of capabilities'
                       % capabilities_id)

    _registered_capabilities[capabilities_id] = caps


def unregister_webapi_capabilities(capabilities_id):
    """Unregister a previously registered set of web API capabilities.

    Args:
        capabilities_id (unicode):
            The unique ID representing a registered collection of capabilities.

    Raises:
        KeyError:
            A set of capabilities matching the ID were not found.
    """
    try:
        del _registered_capabilities[capabilities_id]
    except KeyError:
        logger.error('Failed to unregister unknown web API capabilities '
                     '"%s".',
                     capabilities_id)
        raise KeyError('"%s" is not a registered web API capabilities set'
                       % capabilities_id)



"""A hook for adding entries to the main navigation bar."""

from __future__ import annotations

from djblets.extensions.hooks import ExtensionHook, ExtensionHookPoint


class NavigationBarHook(ExtensionHook, metaclass=ExtensionHookPoint):
    """A hook for adding entries to the main navigation bar.

    This takes a list of entries. Each entry represents something
    on the navigation bar, and is a dictionary with the following keys:

    ``label``:
        The label to display

    ``url``:
        The URL to point to.

    ``url_name``:
        The name of the URL to point to.

    Only one of ``url`` or ``url_name`` is required. ``url_name`` will
    take precedence.

    Optionally, a callable can be passed in for ``is_enabled_for``, which takes
    a single argument (the user) and returns True or False, indicating whether
    the entries should be shown. If this is not passed in, the entries are
    always shown (including for anonymous users).

    If your hook needs to access the template context, it can override
    :py:meth:`get_entries` and return results from there.
    """

    def initialize(self, entries=[], is_enabled_for=None, *args, **kwargs):
        """Initialize the hook.

        This will register each of the entries in the navigation bar.

        Args:
            entries (list of dict):
                The list of dictionary entries representing navigation
                bar items, as documented above.

            is_enabled_for (callable, optional):
                The optional function used to determine if these entries
                should appear for a given page. This is in the format of:

                .. code-block:: python

                   def is_enabled_for(user, request, local_site_name,
                                      **kwargs):
                       return True

                If not provided, the entries will be visible on every page.

            *args (tuple):
                Additional positional arguments. Subclasses should always
                pass these to this class.

            **kwargs (dict):
                Additional keyword arguments. Subclasses should always pass
                these to this class.
        """
        self.entries = entries
        self.is_enabled_for = is_enabled_for

    def get_entries(self, context):
        """Return the navigation bar entries defined in this hook.

        This can be overridden by subclasses if they need more control over
        the entries or need to access the template context.

        Args:
            context (django.template.RequestContext):
                The template context for the page.

        Returns:
            list of dict:
            The list of navigation bar entries. This will be empty if the
            entries are not enabled for this page.
        """
        request = context['request']

        if (not callable(self.is_enabled_for) or
            self.is_enabled_for(user=request.user,
                                request=request,
                                local_site_name=context['local_site_name'])):
            return self.entries
        else:
            return []



"""Unit tests for the actions registry.

Version Added:
    6.0
"""

from __future__ import annotations

from djblets.registries.errors import AlreadyRegisteredError

from reviewboard.actions import AttachmentPoint
from reviewboard.actions.base import ActionPlacement
from reviewboard.actions.tests.base import (
    TestAction,
    TestActionsRegistry,
    TestHeaderAction,
    TestMenuAction,
    TestMenuItemAction,
    TestNested2MenuAction,
    TestNestedMenuAction,
    TooDeeplyNestedAction,
)
from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.testing import TestCase


class ActionsRegistryTests(TestCase):
    """Unit tests for the actions registry.

    Version Added:
        6.0
    """

    def test_register(self) -> None:
        """Testing ActionsRegistry.register"""
        test_action = TestMenuAction()

        actions_registry = TestActionsRegistry()
        actions_registry.register(test_action)

        with self.assertRaises(AlreadyRegisteredError):
            actions_registry.register(test_action)

        self.assertEqual(
            actions_registry._by_attachment_point,
            {
                'review-request': {
                    'menu-action': (
                        test_action,
                        test_action.get_placement('review-request'),
                    ),
                },
            })

        self.assertEqual(actions_registry._deferred_placements, {})

    def test_register_child_of_missing_action(self) -> None:
        """Testing ActionsRegistry.register of action with missing parent
        actions
        """
        test_action = TestMenuItemAction()

        actions_registry = TestActionsRegistry()
        actions_registry.register(test_action)

        with self.assertRaises(AlreadyRegisteredError):
            actions_registry.register(test_action)

        self.assertEqual(actions_registry._by_attachment_point, {})
        self.assertEqual(
            actions_registry._deferred_placements,
            {
                ('menu-action', 'review-request'): {
                    'menu-item-action': test_action,
                },
            })

    def test_register_child_of_missing_placement(self) -> None:
        """Testing ActionsRegistry.register of action with one missing
        placement
        """
        assert TestMenuItemAction.placements

        class MyTestMenuItemAction(TestMenuItemAction):
            placements = [
                *TestMenuItemAction.placements,
                ActionPlacement(attachment='header',
                                parent_id='missing-parent'),
            ]

        test_menu_action = TestMenuAction()
        test_menu_item_action = MyTestMenuItemAction()

        actions_registry = TestActionsRegistry()
        actions_registry.register(test_menu_action)
        actions_registry.register(test_menu_item_action)

        self.assertEqual(
            actions_registry._by_attachment_point,
            {
                'review-request': {
                    'menu-action': (
                        test_menu_action,
                        test_menu_action.get_placement('review-request'),
                    ),
                    'menu-item-action': (
                        test_menu_item_action,
                        test_menu_item_action.get_placement('review-request'),
                    ),
                },
            })

        self.assertEqual(
            actions_registry._deferred_placements,
            {
                ('missing-parent', 'header'): {
                    'menu-item-action': test_menu_item_action,
                },
            })

    def test_register_with_previously_deferred(self) -> None:
        """Testing ActionsRegistry.register of action previously unregistered
        and deferred with children
        """
        test_menu_action = TestMenuAction()
        test_menu_item_action = TestMenuItemAction()

        actions_registry = TestActionsRegistry()
        actions_registry.register(test_menu_action)
        actions_registry.register(test_menu_item_action)

        # Unregister the parent.
        actions_registry.unregister(test_menu_action)

        self.assertEqual(actions_registry._by_attachment_point, {})
        self.assertEqual(
            actions_registry._deferred_placements,
            {
                ('menu-action', 'review-request'): {
                    'menu-item-action': test_menu_item_action,
                },
            })

        # And re-register it.
        actions_registry.register(test_menu_action)

        test_menu_action_placement = \
            test_menu_action.get_placement('review-request')
        test_menu_item_action_placement = \
            test_menu_item_action.get_placement('review-request')

        self.assertIn(test_menu_item_action,
                      test_menu_action_placement.child_actions)
        self.assertIs(test_menu_item_action_placement.parent_action,
                      test_menu_action)

        self.assertEqual(
            actions_registry._by_attachment_point,
            {
                'review-request': {
                    'menu-action': (
                        test_menu_action,
                        test_menu_action_placement,
                    ),
                    'menu-item-action': (
                        test_menu_item_action,
                        test_menu_item_action_placement,
                    ),
                },
            })

        self.assertEqual(actions_registry._deferred_placements, {})

    def test_unregister(self) -> None:
        """Testing ActionsRegistry.unregister"""
        test_action = TestAction()

        actions_registry = TestActionsRegistry()
        actions_registry.register(test_action)

        self.assertEqual(
            actions_registry._by_attachment_point,
            {
                'review-request': {
                    'test': (
                        test_action,
                        test_action.get_placement('review-request'),
                    ),
                },
            })
        self.assertEqual(actions_registry._deferred_placements, {})

        actions_registry.unregister(test_action)

        self.assertEqual(actions_registry._by_attachment_point, {})
        self.assertEqual(actions_registry._deferred_placements, {})

    def test_unregister_with_children(self) -> None:
        """Testing ActionsRegistry.unregister of action with children"""
        test_menu_action = TestMenuAction()
        test_menu_item_action = TestMenuItemAction()

        actions_registry = TestActionsRegistry()
        actions_registry.register(test_menu_action)
        actions_registry.register(test_menu_item_action)

        self.assertEqual(
            actions_registry._by_attachment_point,
            {
                'review-request': {
                    'menu-action': (
                        test_menu_action,
                        test_menu_action.get_placement('review-request'),
                    ),
                    'menu-item-action': (
                        test_menu_item_action,
                        test_menu_item_action.get_placement('review-request'),
                    ),
                },
            })
        self.assertEqual(actions_registry._deferred_placements, {})

        actions_registry.unregister(test_menu_action)

        self.assertEqual(actions_registry._by_attachment_point, {})
        self.assertEqual(
            actions_registry._deferred_placements,
            {
                ('menu-action', 'review-request'): {
                    'menu-item-action': test_menu_item_action,
                },
            })

    def test_get_for_attachment(self) -> None:
        """Testing ActionsRegistry.get_for_attachment"""
        actions_registry = TestActionsRegistry()

        test_action = TestAction()
        test_header_action = TestHeaderAction()

        actions_registry.register(test_action)
        actions_registry.register(test_header_action)

        self.assertEqual(
            actions_registry._by_attachment_point,
            {
                'header': {
                    'header-action': (
                        test_header_action,
                        test_header_action.get_placement('header'),
                    ),
                },
                'review-request': {
                    'test': (
                        test_action,
                        test_action.get_placement('review-request'),
                    ),
                },
            })

        self.assertEqual(
            list(actions_registry.get_for_attachment(AttachmentPoint.HEADER)),
            [
                test_header_action,
            ])

        self.assertEqual(
            list(actions_registry.get_for_attachment(
                AttachmentPoint.REVIEW_REQUEST)),
            [
                test_action,
            ])

    def test_deferred_placements(self) -> None:
        """Testing ActionsRegistry.register with deferred placements"""
        action = TestMenuItemAction()

        # Register the action with an unregistered parent.
        actions_registry = TestActionsRegistry()
        actions_registry.register(action)

        self.assertIs(actions_registry.get_action(action.action_id), action)
        self.assertEqual(
            actions_registry._deferred_placements,
            {
                ('menu-action', 'review-request'): {
                    'menu-item-action': action,
                },
            })

        # Now register the missing action.
        menu_action = TestMenuAction()
        actions_registry.register(menu_action)

        self.assertIn(
            action,
            menu_action.get_placement('review-request').child_actions)

        message = (
            'reviewboard.actions.base.BaseAction.child_actions() is '
            'deprecated and support will be removed in Review Board 10.0. '
            'Please use ActionPlacement.child_actions instead.'
        )

        with self.assertWarns(RemovedInReviewBoard10_0Warning, message):
            self.assertIn(action, menu_action.child_actions)

        self.assertIs(
            actions_registry.get_action(menu_action.action_id),
            menu_action)
        self.assertEqual(actions_registry._deferred_placements, {})

    def test_too_deeply_nested(self) -> None:
        """Testing ActionsRegistry with actions that are too deeply nested"""
        test_menu_action = TestMenuAction()
        nested_action = TestNestedMenuAction()
        nested2_action = TestNested2MenuAction()
        nested3_action = TooDeeplyNestedAction()

        actions_registry = TestActionsRegistry()
        actions_registry.register(test_menu_action)
        actions_registry.register(nested_action)
        actions_registry.register(nested2_action)

        with self.assertLogs() as cm:
            actions_registry.register(nested3_action)

        self.assertEqual(cm.output, [
            'ERROR:reviewboard.actions.registry:Action "nested-3-action" '
            'exceeds the maximum depth limit of 2 for attachment point(s) '
            '"review-request".',
        ])



"""Management command to reset issue counters."""

import logging

from django.core.management.base import BaseCommand, CommandError
from django.utils.translation import gettext as _

from reviewboard.reviews.models import ReviewRequest


class Command(BaseCommand):
    """Management command to reset issue counters."""

    help = 'Resets all calculated issue counts for review requests.'

    def add_arguments(self, parser):
        """Add arguments to the command.

        Args:
            parser (argparse.ArgumentParser):
                The argument parser for the command.
        """
        parser.add_argument(
            'review_request_ids',
            metavar='REVIEW_REQUEST_ID',
            nargs='*',
            help='Specific review request IDs to reset.')

        parser.add_argument(
            '-a',
            '--all',
            action='store_true',
            default=False,
            dest='all',
            help='Reset issue counts for all review requests.')

        parser.add_argument(
            '--recalculate',
            action='store_true',
            default=False,
            dest='recalculate',
            help='Recalculates issue counts for the review requests. '
                 'This is not compatible with --all')

    def handle(self, *args, **options):
        """Handle the command.

        Args:
            *args (tuple):
                Specific review request IDs to reset.

            **options (dict, unused):
                Options parsed on the command line. For this command, no
                options are available.
        """
        update_all = options.get('all')
        recalculate = options.get('recalculate')

        if update_all:
            if recalculate:
                raise CommandError(
                    _('--recalculate cannot be used with --all'))

            q = ReviewRequest.objects.all()
        else:
            pks = []

            for arg in args:
                try:
                    pks.append(int(arg))
                except ValueError:
                    raise CommandError(
                        _('%s is not a valid review request ID')
                        % arg)

            if not pks:
                raise CommandError(
                    _('One or more review request IDs must be provided.'))

            q = ReviewRequest.objects.filter(pk__in=pks)

        q.update(issue_open_count=None,
                 issue_resolved_count=None,
                 issue_dropped_count=None,
                 issue_verifying_count=None)

        if not update_all and recalculate:
            if int(options['verbosity']) > 1:
                root_logger = logging.getLogger('')
                root_logger.setLevel(logging.DEBUG)

            # Load each review request. The issue counters will recalculate,
            # and output will be shown if verbosity > 1.
            list(q)

        if update_all:
            self.stdout.write(_('All issue counts reset.'))
        else:
            self.stdout.write(_('Issue counts for review request(s) %s reset.')
                              % ', '.join(args))



"""Base classes for SSO backends.

Version Added:
    5.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from django.contrib import auth
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.accounts.backends import StandardAuthBackend
from reviewboard.accounts.errors import LoginNotAllowedError

if TYPE_CHECKING:
    from django.contrib.auth.models import User
    from django.http import HttpRequest


class BaseSSOBackend(object):
    """Base class for SSO backends.

    Version Added:
        5.0
    """

    #: The ID of this SSO backend.
    #:
    #: Type:
    #:     str
    backend_id = None

    #: The name of this SSO backend.
    #:
    #: Type:
    #:     str
    name = None

    #: The form used for settings.
    #:
    #: This must be a subclass of
    #: :py:class:`~djblets.siteconfig.forms.SiteSettingsForm`.
    #:
    #: Type:
    #:     type
    settings_form = None

    #: The defaults for siteconfig entries.
    #:
    #: Type:
    #:     dict
    siteconfig_defaults = {}

    #: The text to show in the label for the login button.
    #:
    #: Type:
    #:     str
    login_label = None

    #: The URL to the login flow.
    #:
    #: Type:
    #:     str
    login_url = None

    #: The class type for the view handling the initial login request.
    #:
    #: Type:
    #:     type
    login_view_cls = None

    #: A list of URLs to register.
    #:
    #: Type:
    #:     list
    urls = []

    def __init__(self):
        """Initialize the SSO backend."""

    def is_available(self):
        """Return whether this backend is available.

        Returns:
            tuple:
            A two-tuple. The items in the tuple are:

            1. A bool indicating whether the backend is available.
            2. If the first element is ``False``, this will be a user-visible
               string indicating the reason why the backend is not available.
               If the backend is available, this element will be ``None``.
        """
        raise NotImplementedError

    def is_enabled(self):
        """Return whether this backend is enabled.

        Returns:
            bool:
            ``True`` if this SSO backend is enabled.
        """
        siteconfig = SiteConfiguration.objects.get_current()
        return siteconfig.get('%s_enabled' % self.backend_id, False)

    def login_user(
        self,
        request: HttpRequest,
        user: User,
    ) -> None:
        """Log in the given user.

        Args:
            request (django.http.HttpRequest):
                The HTTP request.

            user (django.contrib.auth.models.User):
                The user to log in.

        Raises:
            reviewboard.accounts.errors.LoginNotAllowedError:
                The user is not allowed to log in.
        """
        if not user.is_active:
            raise LoginNotAllowedError()

        user.backend = '%s.%s' % (StandardAuthBackend.__module__,
                                  StandardAuthBackend.__name__)
        auth.login(request, user)



"""The app definition for reviewboard.reviews.

Version Added:
    6.0
"""

from __future__ import annotations

from django.apps import AppConfig


class ReviewsAppConfig(AppConfig):
    """App configuration for reviewboard.reviews."""

    name = 'reviewboard.reviews'

    def ready(self) -> None:
        """Configure the app once it's ready.

        This will connect signal handlers for the app.

        Version Added:
            6.0
        """
        from reviewboard.reviews.signal_handlers import connect_signal_handlers

        connect_signal_handlers()



"""API resource for managing repositories."""

from __future__ import annotations

from typing import TYPE_CHECKING

from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.db.models import Q
from djblets.registries.errors import ItemLookupError
from djblets.util.decorators import augment_method_from
from djblets.webapi.decorators import (webapi_login_required,
                                       webapi_response_errors,
                                       webapi_request_fields)
from djblets.webapi.errors import (DOES_NOT_EXIST, INVALID_FORM_DATA,
                                   NOT_LOGGED_IN, PERMISSION_DENIED)
from djblets.webapi.fields import (BooleanFieldType,
                                   IntFieldType,
                                   StringFieldType)

from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.scmtools import scmtools_registry
from reviewboard.scmtools.forms import RepositoryForm
from reviewboard.scmtools.models import Repository
from reviewboard.webapi.base import ImportExtraDataError, WebAPIResource
from reviewboard.webapi.decorators import (webapi_check_login_required,
                                           webapi_check_local_site)
from reviewboard.webapi.errors import (BAD_HOST_KEY,
                                       MISSING_REPOSITORY,
                                       MISSING_USER_KEY,
                                       REPO_AUTHENTICATION_ERROR,
                                       REPO_INFO_ERROR,
                                       REPOSITORY_ALREADY_EXISTS,
                                       SERVER_CONFIG_ERROR,
                                       UNVERIFIED_HOST_CERT,
                                       UNVERIFIED_HOST_KEY)
from reviewboard.webapi.mixins import UpdateFormMixin
from reviewboard.webapi.resources import resources

if TYPE_CHECKING:
    from typing import Any

    from django.http import HttpRequest
    from djblets.webapi.resources.base import WebAPIResourceHandlerResult

    from reviewboard.site.models import LocalSite


class RepositoryResource(UpdateFormMixin, WebAPIResource):
    """Provides information on a registered repository.

    Review Board has a list of known repositories, which can be modified
    through the site's administration interface. These repositories contain
    the information needed for Review Board to access the files referenced
    in diffs.
    """

    model = Repository
    form_class = RepositoryForm

    name_plural = 'repositories'

    fields = {
        'id': {
            'type': IntFieldType,
            'description': 'The numeric ID of the repository.',
        },
        'extra_data': {
            'type': dict,
            'description': 'Extra data as part of the repository. Some of '
                           'this will be dependent on the type of repository '
                           'or hosting service being used, and those should '
                           'not be modified. Custom fields can be set by the '
                           'API or extensions.',
            'added_in': '3.0.18',
        },
        'name': {
            'type': StringFieldType,
            'description': 'The name of the repository.',
        },
        'path': {
            'type': StringFieldType,
            'description': 'The main path to the repository, which is used '
                           'for communicating with the repository and '
                           'accessing files.',
        },
        'mirror_path': {
            'type': StringFieldType,
            'description': 'An alternate path to the repository, for '
                           'lookup purposes.',
            'added_in': '1.7.19',
        },
        'visible': {
            'type': BooleanFieldType,
            'description': 'Whether or not this repository is visible (admin '
                           'only).',
            'added_in': '2.0',
        },
        'tool': {
            'type': StringFieldType,
            'description': 'The name of the internal repository '
                           'communication class used to talk to the '
                           'repository. This is generally the type of the '
                           'repository.',
        },
        'bug_tracker': {
            'type': StringFieldType,
            'description': 'The URL to a bug in the bug tracker for '
                           'this repository, with ``%s`` in place of the '
                           'bug ID.',
            'added_in': '2.5',
        },
        'requires_basedir': {
            'type': bool,
            'description': 'Whether the repository requires a patch base '
                           'directory when creating new review requests.',
            'added_in': '4.0.7',
        },
        'requires_change_number': {
            'type': bool,
            'description': 'Whether the repository requires a change number '
                           'when creating new review requests.',
            'added_in': '4.0.7',
        },
        'supports_post_commit': {
            'type': bool,
            'description': 'Whether the repository supports post-commit '
                           'review via the "New Review Request" page.',
            'added_in': '4.0.7',
        },
    }
    uri_object_key = 'repository_id'
    item_child_resources = [
        resources.diff_file_attachment,
        resources.repository_branches,
        resources.repository_commits,
        resources.repository_group,
        resources.repository_info,
        resources.repository_user,
    ]

    allowed_methods = ('GET', 'POST', 'PUT', 'DELETE')

    @webapi_check_login_required
    def get_queryset(self, request, is_list=False, local_site_name=None,
                     show_invisible=False, *args, **kwargs):
        """Returns a queryset for Repository models."""
        local_site = self._get_local_site(local_site_name)

        if is_list:
            queryset = self.model.objects.accessible(
                request.user,
                visible_only=not show_invisible,
                local_site=local_site)

            q = Q()

            if 'q' in request.GET:
                query = request.GET['q']

                if query:
                    q = q & Q(name__istartswith=query)

            if 'name' in request.GET:
                q = q & Q(name__in=request.GET.get('name').split(','))

            if 'path' in request.GET:
                paths = request.GET['path'].split(',')
                q = q & (Q(path__in=paths) | Q(mirror_path__in=paths))

            if 'name-or-path' in request.GET:
                name_or_paths = request.GET['name-or-path'].split(',')
                q = q & (Q(name__in=name_or_paths) |
                         Q(path__in=name_or_paths) |
                         Q(mirror_path__in=name_or_paths))

            if 'tool' in request.GET:
                tool_ids = []

                for tool_name in request.GET['tool'].split(','):
                    try:
                        # Try first as ID, then name.
                        tool = scmtools_registry.get_by_id(tool_name)

                        if tool is not None:
                            tool_ids.append(tool.scmtool_id)
                            continue

                        tool = scmtools_registry.get_by_name(tool_name)

                        if tool is not None:
                            tool_ids.append(tool.scmtool_id)
                    except ItemLookupError:
                        pass

                q = q & Q(scmtool_id__in=tool_ids)

            if 'hosting-service' in request.GET:
                q = q & Q(hosting_account__service_name__in=
                          request.GET['hosting-service'].split(','))

            if 'username' in request.GET:
                usernames = request.GET['username'].split(',')

                q = q & (Q(username__in=usernames) |
                         Q(hosting_account__username__in=usernames))

            queryset = queryset.filter(q).distinct()

            if 'q' in request.GET:
                # If the user is searching for repositories, return them sorted
                # by name.
                queryset = queryset.order_by('name')

            return queryset
        else:
            return self.model.objects.filter(local_site=local_site)

    def parse_tool_field(self, value, **kwargs):
        """Parse the tool field in a request.

        Args:
            value (unicode):
                The name of the tool, as provided in the request.

            **kwargs (dict):
                Additional keyword arguments passed to the method.

        Returns:
            unicode:
            The resulting SCMTool ID to use for the form.

        Raises:
            django.core.exceptions.ValidationError:
                The tool could not be found.
        """
        try:
            return scmtools_registry.get_by_name(name=value).scmtool_id
        except ItemLookupError as e:
            raise ValidationError(e)

    def serialize_tool_field(self, obj, **kwargs):
        """Return the serialized value for the tool field.

        Args:
            obj (reviewboard.scmtools.models.Repository):
                The repository object.

            **kwargs (dict):
                Additional keyword arguments passed to the method.

        Returns:
            str:
            The name of the SCMTool used by this repository.
        """
        return obj.scmtool_class.name

    def serialize_requires_basedir_field(self, obj, **kwargs):
        return not obj.diffs_use_absolute_paths

    def serialize_requires_change_number_field(self, obj, **kwargs):
        return obj.supports_pending_changesets

    def has_access_permissions(self, request, repository, *args, **kwargs):
        return repository.is_accessible_by(request.user)

    def has_modify_permissions(self, request, repository, *args, **kwargs):
        return repository.is_mutable_by(request.user)

    def has_delete_permissions(self, request, repository, *args, **kwargs):
        return repository.is_mutable_by(request.user)

    @webapi_check_login_required
    @webapi_check_local_site
    @webapi_request_fields(
        optional=dict({
            'name': {
                'type': StringFieldType,
                'description': 'Filter repositories by one or more '
                               'comma-separated names.',
                'added_in': '1.7.21',
            },
            'path': {
                'type': StringFieldType,
                'description': 'Filter repositories by one or more '
                               'comma-separated paths or mirror paths.',
                'added_in': '1.7.21',
            },
            'name-or-path': {
                'type': StringFieldType,
                'description': 'Filter repositories by one or more '
                               'comma-separated names, paths, or '
                               'mirror paths.',
                'added_in': '1.7.21',
            },
            'tool': {
                'type': StringFieldType,
                'description': 'Filter repositories by one or more '
                               'comma-separated tool names.',
                'added_in': '1.7.21',
            },
            'hosting-service': {
                'type': StringFieldType,
                'description': 'Filter repositories by one or more '
                               'comma-separated hosting service IDs.',
                'added_in': '1.7.21',
            },
            'username': {
                'type': StringFieldType,
                'description': 'Filter repositories by one or more '
                               'comma-separated usernames.',
                'added_in': '1.7.21',
            },
            'show-invisible': {
                'type': BooleanFieldType,
                'description': 'Whether to list only visible repositories or '
                               'all repositories.',
                'added_in': '2.0',
            },
            'q': {
                'type': StringFieldType,
                'description': 'Filter repositories by those that have a '
                               'name starting with the search term.',
                'added_in': '4.0',
            },
        }, **WebAPIResource.get_list.optional_fields),
        required=WebAPIResource.get_list.required_fields,
        allow_unknown=True
    )
    def get_list(self, request, *args, **kwargs):
        """Retrieves the list of repositories on the server.

        This will only list visible repositories. Any repository that the
        administrator has hidden will be excluded from the list.
        """
        show_invisible = request.GET.get('show-invisible', False)
        return super(RepositoryResource, self).get_list(
            request, show_invisible=show_invisible, *args, **kwargs)

    @augment_method_from(WebAPIResource)
    def get(self, *args, **kwargs):
        """Retrieves information on a particular repository.

        This will only return basic information on the repository.
        Authentication information, hosting details, and repository-specific
        information are not provided.
        """
        pass

    @webapi_check_local_site
    @webapi_login_required
    @webapi_response_errors(BAD_HOST_KEY, INVALID_FORM_DATA, NOT_LOGGED_IN,
                            PERMISSION_DENIED, REPO_AUTHENTICATION_ERROR,
                            REPO_INFO_ERROR, REPOSITORY_ALREADY_EXISTS,
                            SERVER_CONFIG_ERROR, UNVERIFIED_HOST_CERT,
                            UNVERIFIED_HOST_KEY)
    @webapi_request_fields(
        required={
            'name': {
                'type': StringFieldType,
                'description': 'The human-readable name of the repository.',
                'added_in': '1.6',
            },
            'tool': {
                'type': StringFieldType,
                'description': 'The name of the SCMTool to use. Valid '
                               'names can be found in the Administration UI.',
                'added_in': '1.6',
            },
        },
        optional={
            'bug_tracker': {
                'type': StringFieldType,
                'description': 'The URL to a bug in the bug tracker for '
                               'this repository, with ``%s`` in place of the '
                               'bug ID.',
                'added_in': '1.6',
            },
            'bug_tracker_hosting_url': {
                'type': StringFieldType,
                'description': (
                    'The URL to the base of your bug tracker hosting service, '
                    'for services that support this option and when using '
                    '``bug_tracker_type``.'
                ),
                'added_in': '3.0.19',
            },
            'bug_tracker_plan': {
                'type': StringFieldType,
                'description': (
                    'The bug tracker service plan, if specifying '
                    '``bug_tracker_type``. The value is specific to the '
                    'type of bug tracker hosting service.'
                ),
                'added_in': '3.0.19',
            },
            'bug_tracker_type': {
                'type': StringFieldType,
                'description': (
                    'The type of hosting service backing the bug tracker to '
                    'use for this repository, if not specifying '
                    '``bug_tracker``. This may require '
                    '``bug_tracker_hosting_account_username`` and some '
                    'service-specific fields (``bug_tracker_hosting_url``, '
                    '``bug_tracker_plan``, and fields containing the '
                    'service ID).'
                ),
                'added_in': '3.0.19',
            },
            'bug_tracker_use_hosting': {
                'type': BooleanFieldType,
                'description': (
                    "Whether to use the hosting service's own "
                    "bug tracker support for this repository. "
                    "This is dependent on the bug tracker, and "
                    "required ``hosting_type``."
                ),
                'added_in': '3.0.19',
            },
            'encoding': {
                'type': StringFieldType,
                'description': 'The encoding used for files in the '
                               'repository. This is an advanced setting '
                               'and should only be used if you absolutely '
                               'need it.',
                'added_in': '1.6',
            },
            'hosting_account_username': {
                'type': StringFieldType,
                'description': (
                    'The pre-configured username for a hosting service '
                    'account used to connect to this repository. This is '
                    'only used if ``hosting_type`` is set.'
                ),
                'added_in': '3.0.19',
            },
            'hosting_type': {
                'type': StringFieldType,
                'description': (
                    'The type of hosting service backing this repository, if '
                    'not configuring a plain repository. This will require '
                    '``hosting_account_username`` and some service-specific '
                    'fields (``hosting_url``, ``repository_plan``, and '
                    'fields prefixed with the hosting type ID).'
                ),
                'added_in': '3.0.19',
            },
            'hosting_url': {
                'type': StringFieldType,
                'description': (
                    'The URL to the base of your repository hosting service, '
                    'for services that support this option and when using '
                    '``hosting_type``.'
                ),
                'added_in': '3.0.19',
            },
            'mirror_path': {
                'type': StringFieldType,
                'description': 'An alternate path to the repository.',
                'added_in': '1.6',
            },
            'password': {
                'type': StringFieldType,
                'description': 'The password used to access the repository.',
                'added_in': '1.6',
            },
            'path': {
                'type': StringFieldType,
                'description': 'The path to the repository.',
                'added_in': '1.6',
            },
            'public': {
                'type': BooleanFieldType,
                'description': 'Whether or not review requests on the '
                               'repository will be publicly accessible '
                               'by users on the site. The default is true.',
                'added_in': '1.6',
            },
            'raw_file_url': {
                'type': StringFieldType,
                'description': "A URL mask used to check out a particular "
                               "file using HTTP. This is needed for "
                               "repository types that can't access files "
                               "natively. Use ``<revision>`` and "
                               "``<filename>`` in the URL in place of the "
                               "revision and filename parts of the path.",
                'added_in': '1.6',
            },
            'repository_plan': {
                'type': StringFieldType,
                'description': (
                    'The hosting service repository plan, if specifying '
                    '``hosting_type``. The value is specific to the type of '
                    'hosting service.'
                ),
                'added_in': '3.0.19',
            },
            'trust_host': {
                'type': BooleanFieldType,
                'description': 'Whether or not any unknown host key or '
                               'certificate should be accepted. The default '
                               'is false, in which case this will error out '
                               'if encountering an unknown host key or '
                               'certificate.',
                'added_in': '1.6',
            },
            'username': {
                'type': StringFieldType,
                'description': 'The username used to access the repository.',
                'added_in': '1.6',
            },
            'visible': {
                'type': BooleanFieldType,
                'description': 'Whether the repository is visible.',
                'added_in': '2.0',
            },
        },
        allow_unknown=True
    )
    def create(self, request, local_site, parsed_request_fields, *args,
               **kwargs):
        """Creates a repository.

        This will create a new repository that can immediately be used for
        review requests.

        The ``tool`` is a registered SCMTool name. This must be known
        beforehand, and can be looked up in the Review Board administration UI.

        Before saving the new repository, the repository will be checked for
        access. On success, the repository will be created and this will
        return :http:`201`.

        In the event of an access problem (authentication problems,
        bad/unknown SSH key, or unknown certificate), an error will be
        returned and the repository information won't be updated. Pass
        ``trust_host=1`` to approve bad/unknown SSH keys or certificates.
        """
        if not Repository.objects.can_create(request.user, local_site):
            return self.get_no_access_error(request)

        return self._create_or_update(form_data=parsed_request_fields,
                                      request=request,
                                      local_site=local_site,
                                      **kwargs)

    @webapi_check_local_site
    @webapi_login_required
    @webapi_response_errors(DOES_NOT_EXIST, NOT_LOGGED_IN, PERMISSION_DENIED,
                            INVALID_FORM_DATA, SERVER_CONFIG_ERROR,
                            BAD_HOST_KEY, UNVERIFIED_HOST_KEY,
                            UNVERIFIED_HOST_CERT, REPO_AUTHENTICATION_ERROR,
                            REPO_INFO_ERROR)
    @webapi_request_fields(
        optional={
            'archive_name': {
                'type': BooleanFieldType,
                'description': "Whether or not the (non-user-visible) name of "
                               "the repository should be changed so that it "
                               "(probably) won't conflict with any future "
                               "repository names. Starting in 3.0.12, "
                               "performing a DELETE will archive the "
                               "repository, and is the preferred method.",
                'added_in': '1.6.2',
                'deprecated_in': '3.0.12',
            },
            'bug_tracker': {
                'type': StringFieldType,
                'description': 'The URL to a bug in the bug tracker for '
                               'this repository, with ``%s`` in place of the '
                               'bug ID.',
                'added_in': '1.6',
            },
            'bug_tracker_hosting_url': {
                'type': StringFieldType,
                'description': (
                    'The URL to the base of your bug tracker hosting service, '
                    'for services that support this option and when using '
                    '``bug_tracker_type``.'
                ),
                'added_in': '3.0.19',
            },
            'bug_tracker_plan': {
                'type': StringFieldType,
                'description': (
                    'The bug tracker service plan, if specifying '
                    '``bug_tracker_type``. The value is specific to the '
                    'type of bug tracker hosting service.'
                ),
                'added_in': '3.0.19',
            },
            'bug_tracker_type': {
                'type': StringFieldType,
                'description': (
                    'The type of hosting service backing the bug tracker to '
                    'use for this repository, if not specifying '
                    '``bug_tracker``. This may require '
                    '``bug_tracker_hosting_account_username`` and some '
                    'service-specific fields (``bug_tracker_hosting_url``, '
                    '``bug_tracker_plan``, and fields containing the '
                    'service ID).'
                ),
                'added_in': '3.0.19',
            },
            'bug_tracker_use_hosting': {
                'type': BooleanFieldType,
                'description': (
                    "Whether to use the hosting service's own "
                    "bug tracker support for this repository. "
                    "This is dependent on the bug tracker, and "
                    "required ``hosting_type``."
                ),
                'added_in': '3.0.19',
            },
            'encoding': {
                'type': StringFieldType,
                'description': 'The encoding used for files in the '
                               'repository. This is an advanced setting '
                               'and should only be used if you absolutely '
                               'need it.',
                'added_in': '1.6',
            },
            'hosting_account_username': {
                'type': StringFieldType,
                'description': (
                    'The pre-configured username for a hosting service '
                    'account used to connect to this repository. This is '
                    'only used if ``hosting_type`` is set.'
                ),
                'added_in': '3.0.19',
            },
            'hosting_type': {
                'type': StringFieldType,
                'description': (
                    'The type of hosting service backing this repository, if '
                    'not configuring a plain repository. This will require '
                    '``hosting_account_username`` and some service-specific '
                    'fields (``hosting_url``, ``repository_plan``, and '
                    'fields prefixed with the hosting type ID).'
                ),
                'added_in': '3.0.19',
            },
            'hosting_url': {
                'type': StringFieldType,
                'description': (
                    'The URL to the base of your repository hosting service, '
                    'for services that support this option and when using '
                    '``hosting_type``.'
                ),
                'added_in': '3.0.19',
            },
            'mirror_path': {
                'type': StringFieldType,
                'description': 'An alternate path to the repository.',
                'added_in': '1.6',
            },
            'name': {
                'type': StringFieldType,
                'description': 'The human-readable name of the repository.',
                'added_in': '1.6',
            },
            'password': {
                'type': StringFieldType,
                'description': 'The password used to access the repository.',
                'added_in': '1.6',
            },
            'path': {
                'type': StringFieldType,
                'description': 'The path to the repository.',
                'added_in': '1.6',
            },
            'public': {
                'type': BooleanFieldType,
                'description': 'Whether or not review requests on the '
                               'repository will be publicly accessible '
                               'by users on the site. The default is true.',
                'added_in': '1.6',
            },
            'raw_file_url': {
                'type': StringFieldType,
                'description': "A URL mask used to check out a particular "
                               "file using HTTP. This is needed for "
                               "repository types that can't access files "
                               "natively. Use ``<revision>`` and "
                               "``<filename>`` in the URL in place of the "
                               "revision and filename parts of the path.",
                'added_in': '1.6',
            },
            'repository_plan': {
                'type': StringFieldType,
                'description': (
                    'The hosting service repository plan, if specifying '
                    '``hosting_type``. The value is specific to the type of '
                    'hosting service.'
                ),
                'added_in': '3.0.19',
            },
            'trust_host': {
                'type': BooleanFieldType,
                'description': 'Whether or not any unknown host key or '
                               'certificate should be accepted. The default '
                               'is false, in which case this will error out '
                               'if encountering an unknown host key or '
                               'certificate.',
                'added_in': '1.6',
            },
            'username': {
                'type': StringFieldType,
                'description': 'The username used to access the repository.',
                'added_in': '1.6',
            },
            'visible': {
                'type': BooleanFieldType,
                'description': 'Whether the repository is visible.',
                'added_in': '2.0',
            },
        },
        allow_unknown=True
    )
    def update(
        self,
        request: HttpRequest,
        local_site: LocalSite | None,
        parsed_request_fields: dict[str, Any],
        archive_name: (bool | None) = None,
        *args,
        **kwargs,
    ) -> WebAPIResourceHandlerResult:
        """Updates a repository.

        This will update the information on a repository. If the path,
        username, or password has changed, Review Board will try again to
        verify access to the repository.

        In the event of an access problem (authentication problems,
        bad/unknown SSH key, or unknown certificate), an error will be
        returned and the repository information won't be updated. Pass
        ``trust_host=1`` to approve bad/unknown SSH keys or certificates.
        """
        try:
            repository = self.get_object(request, *args, **kwargs)
        except ObjectDoesNotExist:
            return DOES_NOT_EXIST

        if not self.has_modify_permissions(request, repository):
            return self.get_no_access_error(request)

        if archive_name and len(parsed_request_fields) == 1:
            # The only thing we're doing is archiving. We don't want to worry
            # about saving the repository form, because if that fails for any
            # reason (such as an inaccessible repository), we won't be able to
            # archive.
            repository.archive()

            return 200, {
                self.item_result_key: repository,
            }
        else:
            form_data = parsed_request_fields.copy()
            form_data['tool'] = repository.tool.name

            return self._create_or_update(repository=repository,
                                          form_data=form_data,
                                          request=request,
                                          local_site=local_site,
                                          archive=archive_name,
                                          **kwargs)

    @webapi_check_local_site
    @webapi_login_required
    @webapi_response_errors(DOES_NOT_EXIST, NOT_LOGGED_IN, PERMISSION_DENIED)
    def delete(self, request, *args, **kwargs):
        """Deletes a repository.

        Repositories associated with review requests won't be fully deleted
        from the database. Instead, they'll be archived, removing them from
        any lists of repositories but freeing up their name for use in a
        future repository.

        .. versionchanged::
            3.0.12
            Previous releases simply marked a repository as invisible when
            deleting. Starting in 3.0.12, the repository is archived instead.
        """
        try:
            repository = self.get_object(request, *args, **kwargs)
        except ObjectDoesNotExist:
            return DOES_NOT_EXIST

        if not self.has_delete_permissions(request, repository):
            return self.get_no_access_error(request)

        if repository.review_requests.exists():
            # We don't actually delete the repository. We instead archive it.
            # Otherwise, all the review requests are lost. By archiving it,
            # it'll be removed from the UI and from the list in the API.
            repository.archive()
        else:
            repository.delete()

        return 204, {}

    def _create_or_update(
        self,
        form_data: dict[str, Any],
        request: HttpRequest,
        local_site: LocalSite | None,
        repository: (Repository | None) = None,
        archive: (bool | None) = False,
        **kwargs,
    ) -> WebAPIResourceHandlerResult:
        """Create or update a repository.

        Args:
            form_data (dict):
                The repository data to pass to the form.

            request (django.http.HttpRequest):
                The HTTP request from the client.

            local_site (reviewboard.site.models.LocalSite):
                The Local Site being operated on.

            repository (reviewboard.scmtools.models.Repository):
                An existing repository instance to update, if responding to
                a HTTP PUT request.

            archive (bool, optional):
                Whether to archive the repository after updating it.

            **kwargs (dict):
                Keyword arguments representing additional fields handled by
                the API resource.

        Returns:
            djblets.webapi.resources.base.WebAPIResourceHandlerResult:
            The response to send back to the client.
        """
        # Try to determine what we need to set bug_tracker_type to, based on
        # the input and any current settings. We need to do this for two
        # reasons:
        #
        # 1) If the user provides just a `bug_tracker` value, we want to make
        #    sure that the type is set correctly, without the user having to
        #    set it (partially for backwards-compatibility, partially because
        #    it's expected behavior).
        #
        # 2) UpdateFormMixin is going to set a default for `bug_tracker_type`
        #    based on the form field's default value, since it won't find it
        #    on the model.
        if (not form_data.get('bug_tracker_type') and
            (form_data.get('bug_tracker') or
             (repository is not None and
              repository.bug_tracker and
              not repository.extra_data.get('bug_tracker_use_hosting') and
              not repository.extra_data.get('bug_tracker_type')))):
            form_data['bug_tracker_type'] = \
                RepositoryForm.CUSTOM_BUG_TRACKER_ID

        hosting_type = form_data.get('hosting_type')
        hosting_account_id = None

        if hosting_type and hosting_type != 'custom':
            hosting_account_username = \
                form_data.pop('hosting_account_username', None)

            try:
                hosting_account = HostingServiceAccount.objects.get(
                    username=hosting_account_username,
                    hosting_url=form_data.get('hosting_url'),
                    service_name=hosting_type,
                    local_site=local_site)
            except HostingServiceAccount.DoesNotExist:
                return INVALID_FORM_DATA, {
                    'fields': {
                        'hosting_account_username': [
                            'An existing hosting service account with the '
                            'username "%s" could not be found for the hosting '
                            'service "%s".'
                            % (hosting_account_username, hosting_type),
                        ],
                    },
                }

            hosting_account_id = hosting_account.pk

        # Set additional details for the repository form. We want to force
        # some state to be set or unset, and add in any extra form-specific
        # fields.
        form_data.update(kwargs['extra_fields'])
        form_data['hosting_account'] = hosting_account_id

        try:
            return self.handle_form_request(
                data=form_data,
                request=request,
                instance=repository,
                form_kwargs={
                    'limit_to_local_site': local_site,
                    'request': request,
                },
                save_kwargs={
                    'commit': False,
                },
                archive=archive,
                **kwargs)
        except ImportExtraDataError as e:
            return e.error_payload

    def save_form(
        self,
        form: RepositoryForm,
        archive: (bool | None) = None,
        extra_fields: (dict[str, Any] | None) = None,
        **kwargs,
    ) -> Repository:
        """Save the form.

        This will save the repository instance and then optionally archive
        the repository.

        Args:
            form (reviewboard.scmtools.forms.RepositoryForm):
                The form to save.

            archive (bool, optional):
                Whether to archive the repository.

            extra_fields (dict, optional):
                Extra fields from the request not otherwise handled by the
                API resource. Any ``extra_data`` modifications from this will
                be applied to the repository.

            **kwargs (dict):
                Additional keyword arguments to pass to the parent method.

        Returns:
            reviewboard.scmtools.models.Repository:
            The saved repository.
        """
        repository = super().save_form(form=form, **kwargs)

        # Any errors from this will be caught in _create_or_update.
        if extra_fields:
            self.import_extra_data(repository, repository.extra_data,
                                   extra_fields)

        repository.save()
        form.save_m2m()

        if archive:
            repository.archive()

        return repository

    def build_form_error_response(self, form=None, **kwargs):
        """Build an error response based on the form.

        Args:
            form (reviewboard.scmtools.forms.RepositoryForm, optional):
                The repository form.

            **kwargs (dict):
                Keyword arguments passed to this method.

        Returns:
            tuple or django.http.HttpResponse:
            The error response.
        """
        if form is not None:
            if form.get_repository_already_exists():
                return REPOSITORY_ALREADY_EXISTS
            elif form.form_validation_error is not None:
                code = getattr(form.form_validation_error, 'code', None)
                params = getattr(form.form_validation_error, 'params') or {}
                e = params.get('exception')

                if code == 'repository_not_found':
                    return MISSING_REPOSITORY
                elif code == 'repo_auth_failed':
                    return REPO_AUTHENTICATION_ERROR, {
                        'reason': str(e),
                    }
                elif code == 'cert_unverified':
                    cert = e.certificate

                    return UNVERIFIED_HOST_CERT, {
                        'certificate': {
                            'failures': cert.failures,
                            'fingerprint': cert.fingerprint,
                            'hostname': cert.hostname,
                            'issuer': cert.issuer,
                            'valid': {
                                'from': cert.valid_from,
                                'until': cert.valid_until,
                            },
                        },
                    }
                elif code == 'host_key_invalid':
                    return BAD_HOST_KEY, {
                        'hostname': e.hostname,
                        'expected_key': e.raw_expected_key.get_base64(),
                        'key': e.raw_key.get_base64(),
                    }
                elif code == 'host_key_unverified':
                    return UNVERIFIED_HOST_KEY, {
                        'hostname': e.hostname,
                        'key': e.raw_key.get_base64(),
                    }
                elif code in ('accept_cert_failed',
                              'add_host_key_failed',
                              'replace_host_key_failed'):
                    return SERVER_CONFIG_ERROR, {
                        'reason': str(e),
                    }
                elif code == 'missing_ssh_key':
                    return MISSING_USER_KEY
                elif code in ('unexpected_scm_failure',
                              'unexpected_ssh_failure',
                              'unexpected_failure'):
                    return REPO_INFO_ERROR, {
                        'error': str(e),
                    }

        return super(RepositoryResource, self).build_form_error_response(
            form=form,
            **kwargs)


repository_resource = RepositoryResource()



"""The base hosting service class and associated definitions.

This is pending deprecation. Consumers should update their imports to use
the classes in :py:mod:`reviewboard.hostingsvcs.base`.

It includes compatibility imports for:

.. autosummary::
   :nosignatures:

   ~reviewboard.hostingsvcs.base.client.HostingServiceClient
   ~reviewboard.hostingsvcs.base.hosting_service.BaseHostingService
   ~reviewboard.hostingsvcs.base.http.HostingServiceHTTPRequest
   ~reviewboard.hostingsvcs.base.http.HostingServiceHTTPResponse
   ~reviewboard.hostingsvcs.base.repository.RemoteRepository
"""

from __future__ import annotations

import logging

from housekeeping import ClassMovedMixin, func_moved

from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.hostingsvcs.base import client, http
from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService
from reviewboard.hostingsvcs.base.registry import hosting_service_registry
from reviewboard.hostingsvcs.base.registry import HostingServiceRegistry


logger = logging.getLogger(__name__)


class HostingService(ClassMovedMixin,
                     BaseHostingService,
                     warning_cls=RemovedInReviewBoard10_0Warning):
    """An interface to a hosting service for repositories and bug trackers.

    Deprecated:
        8.0:
        This has been moved to :py:class:`reviewboard.hostingsvcs.base.
        hosting_service.BaseHostingService`. The legacy import will be removed
        in Review Board 10.
    """


class HostingServiceClient(ClassMovedMixin,
                           client.HostingServiceClient,
                           warning_cls=RemovedInReviewBoard10_0Warning):
    """Client for communicating with a hosting service's API.

    Deprecated:
        8.0:
        This has been moved to :py:class:`reviewboard.hostingsvcs.base.client.
        HostingServiceClient`. The legacy import will be removed in Review
        Board 10.
    """


class HostingServiceHTTPRequest(ClassMovedMixin,
                                http.HostingServiceHTTPRequest,
                                warning_cls=RemovedInReviewBoard10_0Warning):
    """A request that can use any HTTP method.

    Deprecated:
        8.0:
        This has been moved to :py:class:`reviewboard.hostingsvcs.base.http.
        HostingServiceHTTPRequest`. The legacy import will be removed in Review
        Board 10.
    """


class HostingServiceHTTPResponse(ClassMovedMixin,
                                 http.HostingServiceHTTPResponse,
                                 warning_cls=RemovedInReviewBoard10_0Warning):
    """An HTTP response from the server.

    Deprecated:
        8.0:
        This has been moved to :py:class:`reviewboard.hostingsvcs.base.http.
        HostingServiceHTTPRequest`. The legacy import will be removed in Review
        Board 10.
    """


def get_hosting_services() -> list[type[BaseHostingService]]:
    """Return the list of hosting services.

    Returns:
        list:
        The :py:class:`~reviewboard.hostingsvcs.base.BaseHostingService`
        subclasses.
    """
    return list(hosting_service_registry)


@func_moved(RemovedInReviewBoard10_0Warning,
            new_func=hosting_service_registry.get_hosting_service)
def get_hosting_service(
    name: str,
) -> type[BaseHostingService] | None:
    """Return the hosting service with the given name.

    If the hosting service is not found, None will be returned.

    Args:
        name (str):
            The ID of the hosting service.

    Returns:
        type:
        The hosting service class, or ``None`` if not found.
    """
    return hosting_service_registry.get_hosting_service(name)


@func_moved(RemovedInReviewBoard10_0Warning,
            new_func=hosting_service_registry.register)
def register_hosting_service(
    name: str,
    cls: type[BaseHostingService],
) -> None:
    """Register a custom hosting service class.

    A name can only be registered once. A KeyError will be thrown if attempting
    to register a second time.

    Args:
        name (str):
            The name of the hosting service.

            If the hosting service already has an ID assigned as
            :py:attr:`BaseHostingService.hosting_service_id
            <reviewboard.hostingsvcs.base.BaseHostingService.
            hosting_service_id>`, that value should be passed. Note that this
            will also override any existing ID on the service.

        cls (type):
            The hosting service class.

            This must be a subclass of
            :py:class:`~reviewboard.hostingsvcs.base.BaseHostingService`.
    """
    cls.hosting_service_id = name
    hosting_service_registry.register(cls)


@func_moved(RemovedInReviewBoard10_0Warning,
            new_func=hosting_service_registry.unregister_by_id)
def unregister_hosting_service(
    name: str,
) -> None:
    """Unregister a previously registered hosting service.

    Args:
        name (str):
            The name of the hosting service.
    """
    hosting_service_registry.unregister_by_id(name)


#: Legacy name for HostingServiceHTTPRequest
#:
#: Deprecated:
#:     4.0:
#:     This has been replaced by :py:class:`~reviewboard.hostingsvcs.
#:     base.http.HostingServiceHTTPRequest`.
URLRequest = HostingServiceHTTPRequest


__all__ = [
    'HostingService',
    'HostingServiceClient',
    'HostingServiceHTTPRequest',
    'HostingServiceHTTPResponse',
    'HostingServiceRegistry',
    'URLRequest',
    'get_hosting_service',
    'get_hosting_services',
    'register_hosting_service',
    'unregister_hosting_service',
]

__autodoc_excludes__ = [
    'HostingService',
    'HostingServiceClient',
    'HostingServiceHTTPRequest',
    'HostingServiceHTTPResponse',
    'HostingServiceRegistry',
]



"""Unit tests for the RepositoryUserResource.

Version Added:
    4.0.11
"""

from django.contrib.auth.models import User
from djblets.testing.decorators import add_fixtures
from djblets.webapi.errors import PERMISSION_DENIED
from djblets.webapi.testing.decorators import webapi_test_template

from reviewboard.webapi.errors import INVALID_USER
from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import (repository_user_item_mimetype,
                                                repository_user_list_mimetype)
from reviewboard.webapi.tests.mixins import BasicTestsMetaclass
from reviewboard.webapi.tests.urls import (get_repository_user_item_url,
                                           get_repository_user_list_url,
                                           get_user_item_url)


class ResourceListTests(BaseWebAPITestCase, metaclass=BasicTestsMetaclass):
    """Unit tests for the RepositoryUserResource list API.

    Version Added:
        4.0.11
    """

    fixtures = ['test_users', 'test_scmtools']
    sample_api_url = 'repositories/<id>/users/'
    resource = resources.repository_user
    basic_get_use_admin = True
    basic_post_use_admin = True

    def compare_item(self, item_rsp, user):
        """Compare an item in the results.

        Args:
            item_rsp (dict):
                The response object.

            user (django.contrib.auth.models.User):
                The user to compare against.
        """
        self.assertEqual(item_rsp['id'], user.pk)
        self.assertEqual(item_rsp['username'], user.username)
        self.assertEqual(item_rsp['first_name'], user.first_name)
        self.assertEqual(item_rsp['last_name'], user.last_name)

    def setup_http_not_allowed_item_test(self, user):
        repository = self.create_repository()

        return get_repository_user_list_url(repository)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name,
                             populate_items):
        """Set up data for GET tests.

        Args:
            user (django.contrib.auth.models.User, unused):
                The user for the test.

            with_local_site (bool):
                Whether to set up the test with a local site.

            local_site_name (unicode):
                The name of the local site to use, if appropriate.

            populate_items (bool):
                Whether to populate the list with items.

        Returns:
            tuple:
            A 3-tuple containing the resource URL, the expected mimetype, and
            the list of items.
        """
        repository = self.create_repository(with_local_site=with_local_site)

        if populate_items:
            items = [
                User.objects.get(username='doc'),
                User.objects.get(username='grumpy'),
            ]
            repository.users.add(*items)
        else:
            items = []

        return (get_repository_user_list_url(repository, local_site_name),
                repository_user_list_mimetype,
                items)

    @webapi_test_template
    def test_get_with_no_access(self):
        """Testing the GET <URL> API with Permission Denied"""
        repository = self.create_repository()
        rsp = self.api_get(get_repository_user_list_url(repository),
                           expected_status=403)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], PERMISSION_DENIED.code)

    #
    # HTTP POST tests
    #

    def setup_basic_post_test(self, user, with_local_site, local_site_name,
                              post_valid_data):
        """Set up data for POST tests.

        Args:
            user (django.contrib.auth.models.User, unused):
                The user for the test.

            with_local_site (bool):
                Whether to set up the test with a local site.

            local_site_name (unicode):
                The name of the local site to use, if appropriate.

            post_valid_data (bool):
                Whether to send valid data for the post.

        Returns:
            tuple:
            A 4-tuple containing the resource URL, the expected mimetype, the
            data to post, and the items to use for the tests.
        """
        repository = self.create_repository(with_local_site=with_local_site)

        if post_valid_data:
            post_data = {
                'username': 'doc',
            }
        else:
            post_data = {}

        return (get_repository_user_list_url(repository, local_site_name),
                repository_user_item_mimetype,
                post_data,
                [repository])

    def check_post_result(self, user, rsp, repository):
        """Check the result of a POST operation.

        Args:
            user (django.contrib.auth.models.User, unused):
                The user making the request.

            rsp (dict):
                The request from the API.

            repository (reviewboard.scmtools.models.Repository):
                The repository being changed.
        """
        users = list(repository.users.all())
        self.assertEqual(len(users), 1)
        self.assertEqual(users[0].username, 'doc')
        self.compare_item(rsp['user'], users[0])

    @webapi_test_template
    def test_post_with_no_access(self):
        """Testing the POST <URL> API with Permission Denied"""
        repository = self.create_repository()
        user = User.objects.get(pk=1)

        rsp = self.api_post(
            get_repository_user_list_url(repository),
            {'username': user.username},
            expected_status=403)
        self.assertEqual(rsp['stat'], 'fail')

    @webapi_test_template
    def test_post_with_invalid_user(self):
        """Testing the POST <URL> API with invalid user"""
        self._login_user(admin=True)

        repository = self.create_repository()

        rsp = self.api_post(
            get_repository_user_list_url(repository),
            {'username': 'grabl'},
            expected_status=400)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_USER.code)

    @add_fixtures(['test_site'])
    @webapi_test_template
    def test_post_with_site(self):
        """Testing the POST <URL> API with a local site"""
        self._login_user(admin=True)

        user = User.objects.get(username='doc')

        local_site = self.get_local_site(name=self.local_site_name)
        local_site.users.add(user)

        repository = self.create_repository(with_local_site=True)

        rsp = self.api_post(
            get_repository_user_list_url(repository, self.local_site_name),
            {'username': 'doc'},
            expected_mimetype=repository_user_item_mimetype)
        self.assertEqual(rsp['stat'], 'ok')

        users = list(repository.users.all())
        self.assertEqual(len(users), 1)
        self.assertEqual(users[0], user)

    @add_fixtures(['test_site'])
    @webapi_test_template
    def test_post_with_site_non_member(self):
        """Testing the POST <URL> API with a local site and a user who is not a
        member
        """
        self._login_user(admin=True)

        repository = self.create_repository(with_local_site=True)

        rsp = self.api_post(
            get_repository_user_list_url(repository, self.local_site_name),
            {'username': 'grumpy'},
            expected_status=400)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_USER.code)


class ResourceItemTests(BaseWebAPITestCase, metaclass=BasicTestsMetaclass):
    """Unit tests for the RepositoryUserResource item API."""

    fixtures = ['test_users', 'test_scmtools']
    sample_api_url = 'repositories/<id>/users/<username>/'
    resource = resources.repository_user
    basic_delete_use_admin = True
    basic_get_use_admin = True
    basic_put_use_admin = True

    def setup_http_not_allowed_item_test(self, user):
        """Set up the HTTP not allowed test.

        Args:
            user (django.contrib.auth.models.User, unused):
                The user to set up the test for.

        Returns:
            unicode:
            The URL to fetch for the test.
        """
        repository = self.create_repository()
        doc = User.objects.get(username='doc')
        repository.users.add(doc)

        return get_repository_user_item_url(repository, doc.username)

    def setup_http_not_allowed_list_test(self, user):
        """Set up the HTTP not allowed test.

        Args:
            user (django.contrib.auth.models.User, unused):
                The user to set up the test for.

        Returns:
            str:
            The URL to fetch for the test.
        """
        repository = self.create_repository()
        doc = User.objects.get(username='doc')
        repository.users.add(doc)

        return get_repository_user_item_url(repository, doc.username)

    def compare_item(self, item_rsp, user):
        """Compare an item in the results.

        Args:
            item_rsp (dict):
                The response object.

            user (django.contrib.auth.models.User):
                The user to compare against.
        """
        self.assertEqual(item_rsp['id'], user.pk)
        self.assertEqual(item_rsp['username'], user.username)
        self.assertEqual(item_rsp['first_name'], user.first_name)
        self.assertEqual(item_rsp['last_name'], user.last_name)

    #
    # HTTP DELETE tests
    #

    def setup_basic_delete_test(self, user, with_local_site, local_site_name):
        """Set up data for DELETE tests.

        Args:
            user (django.contrib.auth.models.User, unused):
                The user to run the test as.

            with_local_site (bool):
                Whether to use a local site.

            local_site_name (unicode):
                The name of the local site to test with, if available.

        Returns:
            tuple:
            A 2-tuple containing the resource URL, and a list of objects to
            use.
        """
        repository = self.create_repository(with_local_site=with_local_site)
        doc = User.objects.get(username='doc')
        repository.users.add(doc)

        return (get_repository_user_item_url(repository, doc.username,
                                             local_site_name),
                [repository, doc])

    def check_delete_result(self, user, repository, doc):
        """Check the result of an HTTP DELETE.

        Args:
            user (django.contrib.auth.models.User, unused):
                The user to run the test as.

            repository (reviewboard.scmtools.models.Repository):
                The repository being tested.

            doc (django.contrib.auth.models.User):
                The user being removed from the repository.
        """
        self.assertNotIn(doc, repository.users.all())

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name):
        """Set up data for GET tests.

        Args:
            user (django.contrib.auth.models.User, unused):
                The user for the test.

            with_local_site (bool):
                Whether to set up the test with a local site.

            local_site_name (unicode):
                The name of the local site to use, if appropriate.

        Returns:
            tuple:
            A 3-tuple containing the resource URL, the expected mimetype, and
            the item to compare.
        """
        repository = self.create_repository(with_local_site=with_local_site)
        doc = User.objects.get(username='doc')
        repository.users.add(doc)

        return (get_repository_user_item_url(repository, doc.username,
                                             local_site_name),
                repository_user_item_mimetype,
                doc)

    @webapi_test_template
    def test_get_delete_link(self):
        """Testing the GET <URL> API contains the correct DELETE link"""
        self._login_user(admin=True)

        doc = User.objects.get(username='doc')
        repository = self.create_repository()
        repository.users.add(doc)

        rsp = self.api_get(
            get_repository_user_item_url(repository, doc.username),
            expected_mimetype=repository_user_item_mimetype)

        delete_href = \
            rsp['user']['links']['delete']['href'][len(self.base_url):]

        self.assertEqual(
            delete_href,
            get_repository_user_item_url(repository, doc.username))

        self.assertNotEqual(delete_href, get_user_item_url(doc.username))

    @add_fixtures(['test_site'])
    @webapi_test_template
    def test_get_delete_link_local_site(self):
        """Testing the GET <URL> API contains the correct DELETE link with a local
        site
        """
        self._login_user(admin=True)

        doc = User.objects.get(username='doc')

        local_site = self.get_local_site(self.local_site_name)
        local_site.users.add(doc)

        repository = self.create_repository(local_site=local_site)
        repository.users.add(doc)

        rsp = self.api_get(
            get_repository_user_item_url(repository, doc.username,
                                         local_site.name),
            expected_mimetype=repository_user_item_mimetype)

        delete_href = \
            rsp['user']['links']['delete']['href'][len(self.base_url):]

        self.assertEqual(
            delete_href,
            get_repository_user_item_url(repository, doc.username,
                                         local_site.name))

        self.assertNotEqual(delete_href,
                            get_user_item_url(doc.username, local_site.name))



"""Unit tests for reviewboard.datagrids.columns.ReviewGroupStarColumn.

Version Added:
    8.0
"""

from __future__ import annotations

from django.contrib.auth.models import AnonymousUser
from django.utils.safestring import SafeString

from reviewboard.datagrids.columns import ReviewGroupStarColumn
from reviewboard.datagrids.tests.base import BaseColumnTestCase


class ReviewGroupStarColumnTests(BaseColumnTestCase):
    """Unit tests for reviewboard.datagrids.columns.ReviewGroupStarColumn.

    Version Added:
        8.0
    """

    column = ReviewGroupStarColumn()

    def test_render_data_not_starred(self) -> None:
        """Testing ReviewGroupStarColumn.render_data when not starred"""
        group = self.create_review_group()

        value = self.column.render_data(self.stateful_column, group)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<div class="rb-icon star rb-icon-star-off"'
            ' data-starred="0"'
            ' data-object-type="groups"'
            ' data-object-id="test-group"></div>')

    def test_render_data_starred(self) -> None:
        """Testing ReviewGroupStarColumn.render_data when starred"""
        group = self.create_review_group()

        profile = self.request.user.get_profile()
        profile.star_review_group(group)

        value = self.column.render_data(self.stateful_column, group)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<div class="rb-icon star rb-icon-star-on"'
            ' data-starred="1"'
            ' data-object-type="groups"'
            ' data-object-id="test-group"></div>')

    def test_render_data_as_anonymous(self) -> None:
        """Testing ReviewGroupStarColumn.render_data as anonymous user"""
        group = self.create_review_group()
        self.request.user = AnonymousUser()

        value = self.column.render_data(self.stateful_column, group)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '')

    def test_to_json_not_starred(self) -> None:
        """Testing ReviewGroupStarColumn.to_json when not starred"""
        group = self.create_review_group()

        self.assertIs(self.column.to_json(self.stateful_column, group),
                      False)

    def test_to_json_starred(self) -> None:
        """Testing ReviewGroupStarColumn.to_json when starred"""
        group = self.create_review_group()

        profile = self.request.user.get_profile()
        profile.star_review_group(group)

        self.assertIs(self.column.to_json(self.stateful_column, group),
                      True)
